

def walk_stream(lexer, parser_class, parser_root, listener):
    from antlr4 import CommonTokenStream, ParseTreeWalker
    tokens = CommonTokenStream(lexer)
    parser = parser_class(tokens)
    tree = getattr(parser, parser_root)()
    walker = ParseTreeWalker()
    walker.walk(listener, tree)


def walk_comp(stream):
    from mccode_antlr.grammar import McCompLexer, McCompParser, McCompListener, McCompVisitor

    class Listener(McCompListener):
        def enterProg(self, ctx: McCompParser.ProgContext):
            print("Start the program!")

        def exitProg(self, ctx: McCompParser.ProgContext):
            print("Exit the program!")

    lexer = McCompLexer(stream)
    listener = Listener()  # this needs to be replaced by something that *does* something
    walk_stream(lexer, McCompParser, 'prog', listener)


def walk_instr(stream):
    from mccode_antlr.grammar import McInstrLexer, McInstrParser, McInstrListener, McInstrVisitor

    class Listener(McInstrListener):
        def enterProg(self, ctx: McInstrParser.ProgContext):
            print("Start the program!")

        def exitProg(self, ctx: McInstrParser.ProgContext):
            print("Exit the program!")

    lexer = McInstrLexer(stream)
    listener = Listener()  # this needs to be replaced by something that *does* something
    walk_stream(lexer, McInstrParser, 'prog', listener)


# Press the green button in the gutter to run the script.
if __name__ == '__main__':
    import sys
    from pathlib import Path
    from argparse import ArgumentParser
    from antlr4 import InputStream, FileStream
    parser = ArgumentParser()
    parser.add_argument('-f', '--file', help='file to handle', type=str, default=None)
    parser.add_argument('-t', '--type', help='file type', type=str, default=None)

    args = parser.parse_args()

    file_ext = None if args.type is None else args.type
    file = None if args.file is None else Path(args.file)
    if file is not None and file_ext is None:
        file_ext = file.suffix
    if file is not None and file_ext != file.suffix:
        raise RuntimeError(f"Expected file type {file_ext} does not match provided file extension {file.suffix}!")

    if file is None:
        istream = InputStream(sys.stdin.readline())
    else:
        istream = FileStream(str(file.resolve()), encoding='utf8')

    if 'comp' in file_ext:
        walk_comp(istream)
    elif 'instr' in file_ext:
        walk_instr(istream)
    else:
        raise RuntimeError(f'Unknown file extension type {file_ext}')




from unittest import TestCase
from zenlog import log


def parse_instr_string(instr_source: str):
    from mccode_antlr.loader import parse_mcstas_instr
    return parse_mcstas_instr(instr_source)


def make_mcstas_assembler(name: str):
    from mccode_antlr.assembler import Assembler
    from mccode_antlr.reader import MCSTAS_REGISTRY
    return Assembler(name, registries=[MCSTAS_REGISTRY])


class TestInstr(TestCase):
    def test_parse_empty_trace(self):
        from mccode_antlr.instr import Instr
        from mccode_antlr.common import InstrumentParameter, Expr, Value, DataType
        instr_source = """
        DEFINE INSTRUMENT test_parse(par0=3.14159, double par1 = 49, int par2 =     1010110
            , string par3="this is a long string with spaces", 
            
            string par4, int par5, double par6, par7)
        TRACE
        END
        """
        instr = parse_instr_string(instr_source)
        self.assertTrue(isinstance(instr, Instr))
        self.assertEqual(instr.name, 'test_parse')
        self.assertEqual(len(instr.parameters), 8)
        self.assertEqual(len(instr.components), 0)
        for i, (d_type, val) in enumerate([('double', 3.14159), ('double', 49), ('int', 1010110),
                                           ('string', '"this is a long string with spaces"'),
                                           ('string', None), ('int', None), ('double', None), ('double', None)]):
            p = instr.parameters[i]
            self.assertTrue(isinstance(p, InstrumentParameter))
            self.assertEqual(p.name, f'par{i}')
            self.assertTrue(isinstance(p.value, Expr))
            self.assertEqual(p.value, Value(val, data_type=DataType.from_name(d_type)))

    def test_assemble_empty_trace(self):
        from mccode_antlr.instr import Instr
        from mccode_antlr.common import InstrumentParameter, Expr, Value, DataType

        assembler = make_mcstas_assembler('test_assemble')

        # There are multiple ways to specify instrument parameters:
        # 1. specify a single parameter object to add:
        assembler.parameter(InstrumentParameter('par0', '', Expr.float(3.14159)))
        # 2. specify a single parameter as a string to be parsed:
        assembler.parameter('double par1 = 49')
        # 3. Using the parameters method, specify any number of parameter objects or strings to be parsed (as above)
        #    and/or specify keyword value pairs, where each value can be
        #      a. a two-element list or tuple of 'value', 'unit'
        #      b. a dict with entries 'unit' and 'value'
        #      c. a plain value
        #    in each of these cases, Expr.best converts the provided value if it is not already an Expr object
        assembler.parameters(InstrumentParameter('par2', '', Expr.int(1010110)),
                             'string par3 = "this is a long string with spaces"',
                             InstrumentParameter('par4', '', Expr.str('"the-fourth_parameter"')),
                             'int par5=-9', par6=[Expr.float(None), ''], par7=191, par8=dict(unit='', value='"whoa!"'))

        instr = assembler.instrument
        self.assertTrue(isinstance(instr, Instr))
        self.assertEqual(instr.name, 'test_assemble')
        self.assertEqual(len(instr.parameters), 9)
        self.assertEqual(len(instr.components), 0)
        d_type_vals = [('double', 3.14159), ('double', 49), ('int', 1010110),
                       ('string', '"this is a long string with spaces"'), ('string', '"the-fourth_parameter"'),
                       ('int', -9), ('float', None), ('int', 191), ('string', '"whoa!"')]
        for i, (par, (d_type, val)) in enumerate(zip(instr.parameters, d_type_vals)):
            self.assertTrue(isinstance(par, InstrumentParameter))
            self.assertEqual(par.name, f'par{i}')
            self.assertTrue(isinstance(par.value, Expr))
            self.assertEqual(par.value, Value(val, data_type=DataType.from_name(d_type)))

    def test_assemble_identifier_instance_parameter(self):
        assembler = make_mcstas_assembler('fake_bifrost')

        parameters = [
            'ps1speed/"Hz" = 196', 'ps2speed/"Hz" = 196',
            'fo1speed/"Hz" = 14', 'fo2speed/"Hz" = 14',
            'bw1speed/"Hz" = 14', 'bw2speed/"Hz" = -14',
            'ps1phase/"Hz" = 0', 'ps2phase/"Hz" = 0',
            'fo1phase/"Hz" = 0', 'fo2phase/"Hz" = 0',
            'bw1phase/"Hz" = 0', 'bw2phase/"Hz" = 0',
            'blade0l/"m" = -0.05', 'blade0r/"m" = 0.05',
            'blade1l/"m" = -0.05', 'blade1r/"m" = 0.05',
            'blade2l/"m" = -0.05', 'blade2r/"m" = 0.05',
            'blade3l/"m" = -0.05', 'blade3r/"m" = 0.05',
            'int pulses = 1', 'power/"MW" = 2',
            'source_lambda_min/"angstrom" = 0.75', 'source_lambda_max/"angstrom" = 30',
        ]
        assembler.parameters(*parameters)

        # Mix of initialization temporary variables and calculated instrument parameters.
        pulse_length = 2.86E-3
        assembler.declare(f"""
        double lambda_0;
        double lambda_1;
        double tf_d=1;
        double tf_t={pulse_length / 2};
        double tf_w={pulse_length};
        double tf_tau={1 / 14.0};
        double ps_disk_open_angle=170.0;
        double ps_distance={4.41 + 0.032 + 2.0 - 0.1};
        double bw_disk_open_angle=161.0;
        double bw_distance=78.0;
        """)

        # Chopper-based wavelength calculations can only be done at runtime since they depend on instrument parameters
        new_chopper_calculations = """
        chopper_parameters * pars = calloc(6, sizeof(chopper_parameters));
        pars[0].speed = ps1speed; // Pulse shaping disk 1
        pars[0].phase = ps1phase;
        pars[0].angle = 170.0;
        pars[0].path = ps_distance;
        pars[1].speed = ps2speed; // Pulse shaping disk 2
        pars[1].phase = ps2phase;
        pars[1].angle = pars[0].angle;
        pars[1].path = pars[0].path + 0.02;
        pars[2].speed = fo1speed; // Frame overlap 1
        pars[2].phase = fo1phase;
        pars[2].angle = 38.26;
        pars[2].path = 8.530;
        pars[3].speed = fo2speed; // Frame overlap 2
        pars[3].phase = fo2phase;
        pars[3].angle = 52.01;
        pars[3].path = 14.973;
        pars[4].speed = bw1speed; // Bandwidth disk 1
        pars[4].phase = bw1phase;
        pars[4].angle = 161.0;
        pars[4].path = bw_distance;
        pars[5].speed = bw2speed; // Bandwidth disk 2
        pars[5].phase = bw2phase;
        pars[5].angle = pars[4].angle;
        pars[5].path = pars[4].path + 0.02;

        lambda_0 = source_lambda_min;
        lambda_1 = source_lambda_max;
        double pulse_delay = 2.0e-4; // time to peak brightness after proton pulse T0
        double pulse_width = 2.86e-3; // duration of high flux plateau
        double latest = pulse_delay + pulse_width + 2e-3; // extra for good measure?
        unsigned windows = chopper_wavelength_limits(&source_lambda_min, &source_lambda_max, 6, pars, lambda_0, lambda_1, latest);

        // free the memory used by pars
        if (pars) free(pars);

        if (windows == 0) {
          printf("Chopper train will not pass wavelengths between %f and %f angstrom!\\n", lambda_0, lambda_1);
          printf("Examine the provided chopper speeds and phases.\\n");
        }
        if (windows > 1) {
          printf("Chopper train will pass %u wavelength ranges between %f and %f angstrom\\n", windows, lambda_0, lambda_1);
          printf("but only their envelope between %f and %f angstrom is considered!\\n", source_lambda_min, source_lambda_max);   
        }
        printf("Using source lambda limits %f to %f\\n", source_lambda_min, source_lambda_max);

        // Now calculate the time-focusing window on the first chopper given the wavelength band:
        // ...

        """
        assembler.initialize(new_chopper_calculations)

        # Start the actual instrument:
        assembler.component("Origin", "Progress_bar", at=[0, 0, 0])

        # Elliptic guide gravity initialization and declaration
        def declare_line(name, vec):
            initializer = '{' + ','.join(f'{x}' for x in vec) + '}'
            return f'double {name}[] = {initializer};'

        # Neutron Beam Optical Assembly (NBOA) in-monolith guide:
        nboa = dict()
        nboa['nboa_lens'] = [
            0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.48844444, 0.06121889, 0.02349,
            0.40373556, 0.07626444, 0.015, 0.39671
        ]
        nboa['nboa_vert'] = [3.5, 3.5, 3.0, 3.0, 2.5, 2.5, 2.5, 2.5, 0.0, 2.0, 2.0, 0.0, 2.0]
        nboa['nboa_horz'] = [3.0, 3.0, 2.5, 2.5, 2.5, 2.0, 1.5, 1.5, 0.0, 1.5, 1.5, 0.0, 1.5]

        assembler.declare('\n'.join([declare_line(n, v) for n, v in nboa.items()]), source=__file__)

        el6_lens = nboa['nboa_lens']

        guide = assembler.component("NBOA", 'Elliptic_guide_gravity', at=([0, 0, 1e-5], "PREVIOUS"))
        # make sure that array-valued parameters can be set:
        guide.set_parameters(dimensionsAt="\"mid\"", R0=0.99, Qc=0.0217, alpha=3.1, W=0.003, xwidth=0.069634,
                             yheight=0.04862, linxw=3.4578, loutxw=0.415155, linyh=1.36, loutyh=3.487681,
                             seglength="nboa_lens", mvaluesright="nboa_horz", mvaluesleft="nboa_horz",
                             mvaluestop="nboa_vert", mvaluesbottom="nboa_vert", nSegments=len(el6_lens),
                             l=sum(el6_lens))

        # Bandwidth chopper
        bw_upstream_gap = 0.1
        bwc1 = assembler.component("BWC1", "DiskChopper", at=([0, 0, bw_upstream_gap], "PREVIOUS"))
        bwc1.set_parameters(theta_0="bw_disk_open_angle", radius='1+2+3+4', nslit='(2+4)/(1+2)', yheight=0.1,
                            nu="bw1speed", phase="bw1phase/2")

        for par in ('theta_0', 'nslit', 'nu', 'phase'):
            self.assertFalse(bwc1.get_parameter(par).value.is_constant)
        for par in ('radius', 'yheight'):
            self.assertTrue(bwc1.get_parameter(par).value.is_constant)

        # nslit is not a constant parameter because the expression parsing does not simplify grouped expressions
        nslit = bwc1.get_parameter('nslit').value.simplify()
        # But explicitly simplifying the expression is possible (in this simple case) and yields a constant value
        self.assertTrue(nslit.simplify().is_constant)

        phase = bwc1.get_parameter('phase').value  # == bw1phase / 2
        # In contrast, phase is Expr(BinaryOp('/', Expr.id('bw1phase'), Expr.int(2))) which can't be simplified
        self.assertFalse(phase.simplify().is_constant)

        # But we can attempt to parse the declarations and instantiation blocks from the instrument
        from mccode_antlr.translators.c_listener import extract_c_declared_expressions, evaluate_c_defined_expressions
        variables = {x.name:  x.value for x in assembler.instrument.parameters}
        for dec in assembler.instrument.declare:
            variables.update(extract_c_declared_expressions(dec.source))

        # defined as
        # TODO this does not work because the simple "C"-style expression parser doesn't know about pointers
        try:
            for init in assembler.instrument.initialize:
                variables = evaluate_c_defined_expressions(variables, init.source)
        except AttributeError:
            log.warn('Evaluating INITIALIZE section failed; see preceding errors for hints for why. '
                     'This is not a test failure condition (for now). Continuing')

        evaluated_phase = phase.evaluate(variables)
        self.assertTrue(evaluated_phase.is_constant)

        theta_0 = bwc1.get_parameter('theta_0').value
        self.assertTrue(theta_0.evaluate(variables).is_constant)

        # We want to know whether an expression involves any instrument parameters
        # -- which phase and nu do, but others do not:
        for x in ('phase', 'nu'):
            par = bwc1.get_parameter(x).value
            dep = any(par.depends_on(inst_par.name) for inst_par in assembler.instrument.parameters)
            self.assertTrue(dep)
        for x in ('theta_0', 'radius', 'nslit', 'yheight'):
            par = bwc1.get_parameter(x).value
            self.assertFalse(any(par.depends_on(inst_par.name) for inst_par in assembler.instrument.parameters))

    def assertRotationsEqual(self, r1, r2):
        from mccode_antlr.instr.orientation import Rotation, Matrix
        self.assertTrue(isinstance(r1, Rotation))
        self.assertTrue(isinstance(r2, Rotation))
        d = r1 - r2
        self.assertTrue(isinstance(d, Matrix))
        self.assertEqual(round(abs(d), 14), Matrix())

    def _positioning_evaluator(self, instr):
        """Common positioning checks for `test_read_positioning` and `test_assemble_positioning`"""
        from mccode_antlr.common import Expr
        from mccode_antlr.instr.orientation import Vector, Rotation
        z, o = Expr.float(0), Expr.float(1)

        left = instr.get_component('left')
        v3 = Vector(z, z, o)
        self.assertEqual(v3, left.orientation.position())
        self.assertEqual(Rotation(z, z, -o, z, o, z, o, z, z), left.orientation.rotation())
        self.assertEqual(left.orientation.rotation(), left.orientation.rotation('coordinates').inverse())

        up = instr.get_component('up')
        v4 = Vector(o, z, z)
        self.assertEqual(v4, up.orientation.position() - left.orientation.position())
        up_orientation = Rotation(z, z, -o, -o, z, z, z, o, z)
        self.assertRotationsEqual(up_orientation, up.orientation.rotation())
        self.assertEqual(Vector(o, z, o), up.orientation.position())
        self.assertRotationsEqual(up_orientation, up.orientation.rotation('coordinates').inverse())

        last = instr.get_component('last')
        v5 = Vector(z, o, z)
        self.assertRotationsEqual(up_orientation, last.orientation.rotation())
        self.assertEqual(v5, last.orientation.position() - up.orientation.position())

    def test_assemble_positioning(self):
        """Equivalent test to `test_read_positioning` but using an assembled instrument"""
        assembler = make_mcstas_assembler('orientation_test')
        origin = assembler.component("origin", "Progress_bar", at=[0, 0, 0])
        left = assembler.component('left', 'Arm', at=([0, 0, 1], origin), rotate=[0, 90, 0])
        up = assembler.component("up", 'Arm', at=([0, 0, 1], left), rotate=[-90, 0, 0])
        assembler.component("last", 'Arm', at=([0, 0, 1], up), rotate=[0, 0, 0])
        self._positioning_evaluator(assembler.instrument)

    def test_read_positioning(self):
        """Equivalent test to `test_assemble_positioning` but using a parsed instrument"""
        instr_source = """DEFINE INSTRUMENT orientation_test() TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT left = Arm() AT (0, 0, 1) RELATIVE PREVIOUS ROTATED (0, 90, 0) RELATIVE origin
        COMPONENT up = Arm() AT (0, 0, 1) RELATIVE left ROTATED (-90, 0, 0) RELATIVE left
        COMPONENT last = Arm() AT (0, 0, 1) RELATIVE up
        END """
        self._positioning_evaluator(parse_instr_string(instr_source))

    def _simple_position_tests(self, instr, positions: dict):
        from mccode_antlr.common import Expr
        from mccode_antlr.instr.orientation import Vector
        positions = {k: Vector(*[Expr.float(x) for x in v]) for k, v in positions.items()}
        for instance in instr.components:
            self.assertEqual(positions[instance.name], instance.orientation.position())
            self.assertEqual(positions[instance.name], instance.orientation.position('axes'))
            self.assertEqual(positions[instance.name], instance.orientation.position('coordinates'))

    def test_simple_positioning(self):

        from math import pi, cos, sin
        instr_source = """DEFINE INSTRUMENT simple_test() TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT slit = Arm() at (0, 0, 10) RELATIVE origin
        COMPONENT sample = Arm() at (0, 0, 10) RELATIVE slit
        END"""
        instr = parse_instr_string(instr_source)
        positions = {'origin': (0, 0, 0), 'slit': (0, 0, 10), 'sample': (0, 0, 20)}
        self._simple_position_tests(instr, positions)

        instr_source = """DEFINE INSTRUMENT slightly_more_complex() TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT guide_start = Arm() AT (0.01277, 0, 1.930338) RELATIVE origin ROTATED (0, -0.56, 0) RELATIVE origin
        COMPONENT guide = Arm() AT (0, 0, 0) RELATIVE guide_start
        COMPONENT guide_end = Arm() AT (0, 0, 4.33) RELATIVE guide
        END
        """
        instr = parse_instr_string(instr_source)
        positions = {'origin': (0, 0, 0), 'guide_start': (0.01277, 0, 1.930338), 'guide': (0.01277, 0, 1.930338),
                     'guide_end': (0.01277 - 4.33*sin(pi/180*0.56), 0, 1.930338 + 4.33*cos(pi/180*0.56))}
        self._simple_position_tests(instr, positions)
        pos_hat = (0.006615, 0, 0.999978)
        pos = instr.components[2].orientation.position()
        distance = pos.length()
        vector = pos / distance
        self.assertAlmostEqual(vector.x, pos_hat[0], 6)
        self.assertAlmostEqual(vector.y, pos_hat[1], 6)
        self.assertAlmostEqual(vector.z, pos_hat[2], 6)
        self.assertAlmostEqual(distance, 1.930380239005777, 6)

        # Combining the position and rotation (reduced) operations should yield the same position
        comb = instr.components[2].orientation.combine().reduce()
        self.assertEqual(pos, comb.position())

    def test_copy(self):
        from mccode_antlr.instr import Instr
        from mccode_antlr.common import InstrumentParameter, Expr, Value, DataType
        instr_source = """
        DEFINE INSTRUMENT test_copy(par0=3.14159, double par1 = 49, int par2 =     1010110
            , string par3="this is a long string with spaces", 
            
            string par4, int par5, double par6, par7)
        TRACE
        COMPONENT first = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT second = Arm() AT (0, 0, 1) RELATIVE first ROTATED (0, 90, 0) RELATIVE first
        END
        """
        instr = parse_instr_string(instr_source)
        instr_copy = instr.copy()
        self.assertTrue(isinstance(instr_copy, Instr))
        self.assertEqual(instr.name, instr_copy.name)
        self.assertEqual(len(instr.parameters), len(instr_copy.parameters))
        self.assertEqual(len(instr.components), len(instr_copy.components))
        for p, p_copy in zip(instr.parameters, instr_copy.parameters):
            self.assertTrue(isinstance(p_copy, InstrumentParameter))
            self.assertEqual(p.name, p_copy.name)
            self.assertEqual(p.value, p_copy.value)

        instr_copy.add_parameter(InstrumentParameter('par8', '', Expr.float(1.6189)))
        self.assertEqual(len(instr.parameters), len(instr_copy.parameters) - 1)

    def test_mcpl_split(self):
        from mccode_antlr.instr import Instr
        instr_source = """
        DEFINE INSTRUMENT test_copy(par0=3.14159, double par1 = 49, int par2 =     1010110
            , string par3="this is a long string with spaces", 

            string par4, int par5, double par6, par7)
        TRACE
        COMPONENT first = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT second = Arm() AT (0, 0, 1) RELATIVE first ROTATED (0, 90, 0) RELATIVE first
        END
        """
        instr = parse_instr_string(instr_source)
        before, after = instr.mcpl_split('first', filename='test_mcpl_split')
        self.assertTrue(isinstance(before, Instr))
        self.assertTrue(isinstance(after, Instr))
        self.assertEqual(len(before.components), 1)
        self.assertEqual(len(after.components), 2)
        self.assertEqual(before.components[0].name, 'first')
        self.assertEqual(after.components[0].name, 'first')
        self.assertEqual(after.components[1].name, 'second')
        self.assertEqual(before.components[0].type.name, 'MCPL_output')
        self.assertEqual(after.components[0].type.name, 'MCPL_input')

    def test_tas1_c1(self):
        from mccode_antlr.loader.loader import parse_mccode_instr_parameters
        contents = """ DEFINE INSTRUMENT tas(PHM=-37.077,TTM=-74,C1=30) TRACE END"""
        instr_parameters = parse_mccode_instr_parameters(contents)
        self.assertEqual(len(instr_parameters), 3)
        self.assertEqual(instr_parameters[0].name, 'PHM')
        self.assertEqual(instr_parameters[1].name, 'TTM')
        self.assertEqual(instr_parameters[2].name, 'C1')
        self.assertEqual(instr_parameters[0].value, -37.077)
        self.assertEqual(instr_parameters[1].value, -74)
        self.assertEqual(instr_parameters[2].value, 30)

    def test_used_parameter_check(self):
        from mccode_antlr.assembler import Assembler
        from mccode_antlr.reader import MCSTAS_REGISTRY
        assembler = Assembler('test_used_parameter_check', registries=[MCSTAS_REGISTRY])
        assembler.parameter('double par0/"pi" = 3.14159')
        assembler.parameter('int par1 = 49')
        assembler.parameter('int par2 = 1010110')
        assembler.parameter('string par3 = "this is a long string with spaces"')
        assembler.parameter('string par4 = "the-fourth_parameter"')
        assembler.parameter('par5 = -9.11')
        assembler.parameter('a4/"degree" = 90')
        assembler.component('origin', 'Progress_bar', at=(0, 0, 0))
        assembler.component('left', 'PSD_monitor', at=(0, 0, 1), parameters=dict(
            xwidth='par0', yheight='2*fmod(par5, 0.1)', nx='par1', ny='par2 >> 4', restore_neutron=1))
        # TODO (maybe) the right shift operator is not implemented, instead the expression is parsed as
        #      BinOp('>', [BinOp('>', ['par2', '']), '4']) {or similar} which passes this test but may not
        #      be sufficient for all cases
        assembler.component('up', 'PSD_monitor', at=(0, 0, 2), parameters=dict(
            xwidth='cos(DEG2RAD * a4)', yheight='sin(DEG2RAD * a4)', nx='-a4 * DEG2RAD', restore_neutron=1))

        inst = assembler.instrument
        self.assertEqual(7, len(inst.parameters))
        self.assertEqual(3, len(inst.components))
        self.assertTrue(inst.parameter_used('par0'))
        self.assertTrue(inst.parameter_used('par1'))
        self.assertTrue(inst.parameter_used('par2'))
        self.assertFalse(inst.parameter_used('par3'))
        self.assertFalse(inst.parameter_used('par4'))
        self.assertTrue(inst.parameter_used('par5'))
        self.assertTrue(inst.parameter_used('a4'))
        self.assertEqual(2, inst.check_instrument_parameters(), )

    def test_split_broken_reference(self):
        from textwrap import dedent
        from mccode_antlr.instr import Instance
        from mccode_antlr.common.expression import Expr
        from mccode_antlr.instr.orientation import Vector
        instr = dedent("""\
        DEFINE INSTRUMENT test_tof(phase/"degree"=0)
        TRACE
        COMPONENT Origin = Progress_bar() AT (0, 0, 0) ABSOLUTE 
        COMPONENT ESS_source = ESS_butterfly(
            sector="W", beamline=4, yheight=0.03, cold_frac=0.5, dist=1, focus_xw=0.01, focus_yh=0.01, 
            c_performance=1.0, t_performance=1.0, Lmin=1, Lmax=2, tmax_multiplier=1.5, n_pulses=1, acc_power=2.0
        ) AT (0, 0, 0) ABSOLUTE 
        COMPONENT guide = Guide_gravity(w1=0.01, w2=0.05, h1=0.01, h2=0.05, l=8.0, m=3.5, G=-9.82) AT (0, 0, 1) ABSOLUTE 
        COMPONENT guide_end = Arm() AT (0, 0, 8.0) RELATIVE guide
        COMPONENT chopper = DiskChopper(radius=0.35, nu=14, phase=phase, theta_0=115) AT (0, 0, 0.01) RELATIVE guide_end
        COMPONENT split_at = Arm() AT (0, 0, 1e-08) RELATIVE chopper
        COMPONENT sample = Incoherent(
            radius=0.005, yheight=0.02, thickness=0.001, focus_ah=2.0, focus_aw=2.0, target_x=1.0, target_y=0.0, 
            target_z=0.0, Etrans=0.0, deltaE=0.2
        ) AT (0, 0, 1) RELATIVE chopper 
        END""")
        instr = parse_instr_string(instr)
        first, second = instr.split('split_at', remove_unused_parameters=True)
        # The sample _should not_ still depend on the chopper, which is only present in the first instrument!
        sample = second.components[-1]
        self.assertEqual(sample.name, 'sample')
        at_ref = sample.at_relative
        self.assertTrue(isinstance(at_ref[0], Vector))
        self.assertFalse(isinstance(at_ref[1], Instance))
        self.assertTrue(at_ref[1] is None)
        # the position of the sample depends on the source, guide, guide-end, and chopper
        v = Vector(Expr.float(0), Expr.float(0), Expr.float(1 + 8 + 0.01 + 1))
        self.assertEqual(v, at_ref[0])


class CompiledTest(TestCase):
    def setUp(self):
        import subprocess
        from mccode_antlr.config import config
        try:
            subprocess.run([config['cc'].get(str), '--version'], check=True)
        except FileNotFoundError:
            log.info(f'Provide alternate C compiler via MCCODE_ANTLR_CC environment variable')
            self.skipTest(f"C compiler {config['cc']} not found")


class CompiledInstr(CompiledTest):
    def _compile_and_run(self, instr, parameters, run=True):
        from mccode_antlr.compiler.c import compile_instrument, CBinaryTarget, run_compiled_instrument
        from mccode_antlr.translators.target import MCSTAS_GENERATOR
        from mccode_antlr.loader import read_mccode_dat
        from mccode_antlr.config import config as module_config
        from tempfile import TemporaryDirectory
        from os import R_OK, access
        from pathlib import Path

        target = CBinaryTarget(mpi=False, acc=False, count=1, nexus=False)
        config = dict(default_main=True, enable_trace=False, portable=False, include_runtime=True,
                      embed_instrument_file=False, verbose=False)

        with TemporaryDirectory() as directory:
            try:
                compile_instrument(instr, target, directory, generator=MCSTAS_GENERATOR, config=config, dump_source=True)
            except RuntimeError as e:
                log.error(f'Failed to compile instrument: {e}')
                raise e
            binary = Path(directory).joinpath(f'{instr.name}{module_config["ext"].get(str)}')
            self.assertTrue(binary.exists())
            self.assertTrue(binary.is_file())
            self.assertTrue(access(binary, R_OK))
            if run:
                run_compiled_instrument(binary, target, f"--dir {directory}/instr {parameters}")
                sim_files = list(Path(directory).glob('**/*.dat'))
                dats = {file.stem: read_mccode_dat(file) for file in sim_files}
                return dats
            return None

    def test_one_axis(self):
        from math import pi, asin, sqrt
        from mccode_antlr.loader import parse_mcstas_instr
        d_spacing = 3.355  # (002) for Highly-ordered Pyrolytic Graphite
        mean_energy = 5.0
        energy_width = 0.5
        mean_ki = sqrt(mean_energy / 2.0722)
        instr = f"""
        DEFINE INSTRUMENT splitRunTest(a1=0, a2=0, virtual_source_x=0.05, virtual_source_y=0.1, string newname)
        TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT source = Source_simple(yheight=2*virtual_source_y, xwidth=0.2, dist=1.5, focus_xw=0.06, focus_yh=0.12,
                                         E0={mean_energy}, dE={energy_width})
                           AT (0, 0, 0) RELATIVE origin
        COMPONENT m0 = PSD_monitor(xwidth=0.1, yheight=0.15, nx=100, ny=160, restore_neutron=1) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT guide = Guide_gravity(w1 = 0.06, h1 = 0.12, w2 = 0.05, h2 = 0.1, l = 30, m = 4) 
                          AT (0, 0, 1.5) RELATIVE  PREVIOUS
        COMPONENT guide_end = Arm() AT (0, 0, 30) RELATIVE PREVIOUS
        COMPONENT m1 = PSD_monitor(xwidth=0.1, yheight=0.15, nx=100, ny=160, restore_neutron=1) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT aperture = Slit(xwidth=virtual_source_x, yheight=virtual_source_y) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT split_at = Arm() AT (0, 0, 0.0001) RELATIVE PREVIOUS
        COMPONENT m3 = PSD_monitor(xwidth=0.1, yheight=0.15, nx=100, ny=160, restore_neutron=1) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT mono_point = Arm() AT (0, 0, 0.8) RELATIVE split_at
        METADATA "txt" "something" %{{
            This is some unparsed metadata that will be included as a literal string in the instrument.
        %}}
        COMPONENT mono = Monochromator_curved(zwidth = 0.02, yheight = 0.02, NH = 13, NV = 7, DM={d_spacing}) 
                         AT (0, 0, 0) RELATIVE  mono_point ROTATED (0, a1, 0) RELATIVE mono_point
        COMPONENT sample_arm = Arm() AT (0, 0, 0) RELATIVE mono_point ROTATED (0, a2, 0) RELATIVE mono_point
        COMPONENT detector = Monitor(xwidth=0.1, yheight=0.15, restore_neutron=1) AT (0, 0, 0.8) RELATIVE sample_arm
        COMPONENT lmon = L_monitor(filename=newname) AT (0, 0, 0.001) RELATIVE PREVIOUS
        END
        """
        instr = parse_mcstas_instr(instr)

        a1 = asin(pi / d_spacing / mean_ki) * 180 / pi
        self.assertAlmostEqual(a1, 37.0722, 4)
        parameters = f'a1={a1} a2={2 * a1} -n 1000000'

        dats = self._compile_and_run(instr, parameters)
        self.assertEqual(len(dats), 5)
        self.assertEqual(dats['m0'].data.shape, (3, 160, 100))
        self.assertEqual(dats['m1'].data.shape, (3, 160, 100))
        self.assertEqual(dats['m3'].data.shape, (3, 160, 100))
        self.assertEqual(dats['detector'].data.shape, (3, ))

        # Moving farther from the source means less (but finite) intensity in equivalent monitors
        self.assertTrue(sum(sum(dats['m0']['I'])) > sum(sum(dats['m1']['I'])) > sum(sum(dats['m3']['I'])) > 0)
        # The detector has been positioned correctly to collect intensity
        self.assertTrue(dats['detector']['I'] > 0)

    def test_assembled_parameters(self):
        """Check that setting an instance parameter to a value that is an instrument parameter name works"""
        assembler = make_mcstas_assembler('assembled_parameters_test_instr')
        assembler.parameter("double par0 = 3.14159")
        origin = assembler.component("origin", "Progress_bar", at=[0, 0, 0])
        left = assembler.component('left', 'Slit', at=([0, 0, 1], origin), rotate=[0, 90, 0],
                                   parameters=dict(xwidth='par0', yheight='2*fmod(par0, 0.1)'))

        self._compile_and_run(assembler.instrument, None, run=False)

    def test_vector_component_parameter(self):
        """Some components can use vector parameters, which must be initialized by initializer lists"""
        from mccode_antlr.common import Value
        # TODO Update the registry with the new Conics_* components, then update this test (if necessary)
        instr = """
        DEFINE INSTRUMENT test_vector_component_parameter() TRACE
        COMPONENT cEH = Conics_EH(R0=0.99, alpha=6.07, W=0.003, m=3, nshells=4, focal_length_u=10, focal_length_d=10,
                                  radii={0.05236, 0.03, 0.01, 0.0031416}, le=0.25, lh=0.25, disk=1)
        AT (0,0,0) ABSOLUTE
        COMPONENT cEH2 = Conics_EH(R0=0.99, alpha=6.07, W=0.003, m=3, nshells=0, focal_length_u=10, focal_length_d=10,
                                   le=0.25, lh=0.25, disk=0)
        AT (0,0,1) ABSOLUTE
        END
        """
        instr = parse_instr_string(instr)
        self.assertEqual(instr.components[0].get_parameter('radii').value, Value([0.05236, 0.03, 0.01, 0.0031416]))

        try:
            self._compile_and_run(instr, '-n 1000', run=True)
        except RuntimeError as e:
            log.error(f'Failed to compile instrument: {e}')
            self.fail(f'Failed to compile instrument {e}')

    def test_split_broken_reference_compiles(self):
        from textwrap import dedent
        instr = dedent("""\
        DEFINE INSTRUMENT test_tof(phase/"degree"=0, ang/"degree"=0)
        TRACE
        COMPONENT Origin = Progress_bar() AT (0, 0, 0) ABSOLUTE 
        COMPONENT ESS_source = ESS_butterfly(
            sector="W", beamline=4, yheight=0.03, cold_frac=0.5, dist=1, focus_xw=0.01, focus_yh=0.01, 
            c_performance=1.0, t_performance=1.0, Lmin=1, Lmax=2, tmax_multiplier=1.5, n_pulses=1, acc_power=2.0
        ) AT (0, 0, 0) ABSOLUTE 
        COMPONENT guide = Guide_gravity(w1=0.01, w2=0.05, h1=0.01, h2=0.05, l=8.0, m=3.5, G=-9.82) AT (0, 0, 1) ABSOLUTE 
        COMPONENT guide_end = Arm() AT (0, 0, 8.0) RELATIVE guide
        COMPONENT chopper = DiskChopper(radius=0.35, nu=14, phase=phase, theta_0=115) AT (0, 0, 0.01) RELATIVE guide_end
        COMPONENT rotate = Arm() AT (0, 0, 0) RELATIVE chopper ROTATED (0, ang, 0) RELATIVE chopper
        COMPONENT aperture = Slit(xwidth=0.05, yheight=0.05) AT (0, 0, 1) RELATIVE rotate
        COMPONENT split_at = Arm() AT (0, 0, 1e-08) RELATIVE rotate
        COMPONENT sample = Incoherent(
            radius=0.005, yheight=0.02, thickness=0.001, focus_ah=2.0, focus_aw=2.0, target_x=1.0, target_y=0.0, 
            target_z=0.0, Etrans=0.0, deltaE=0.2
        ) AT (0, 0, 1) RELATIVE aperture
        END""")
        instr = parse_instr_string(instr)
        _, second = instr.split('split_at', remove_unused_parameters=True)
        try:
            self._compile_and_run(second, '-n 1000 ang=10', run=True)
        except RuntimeError as e:
            log.error(f'Failed to compile instrument: {e}')
            self.fail(f'Failed to compile instrument {e}')



class CompiledMCPL(CompiledTest):
    def setUp(self):
        import subprocess
        try:
            subprocess.run(['mcpl-config', '--version'], check=True)
        except FileNotFoundError:
            self.skipTest('mcpl-config not found')

    def test_mcpl_split_run(self):
        # Adapted from Test_MCPL_*.instr in ${MCCODE}/mcstas-comps/examples
        from mccode_antlr.instr import Instr
        from mccode_antlr.common import ComponentParameter, Expr
        from mccode_antlr.compiler.c import compile_instrument, run_compiled_instrument, CBinaryTarget
        from mccode_antlr.translators.target import MCSTAS_GENERATOR
        from mccode_antlr.loader import read_mccode_dat
        from tempfile import TemporaryDirectory
        from os import R_OK, access
        from pathlib import Path
        from random import randint
        from numpy import allclose

        instr_source = """
        DEFINE INSTRUMENT Test_MCPL_output()
        USERVARS %{ double flag; %}
        TRACE
        COMPONENT Origin = Progress_bar() AT (0, 0, 0) ABSOLUTE
        COMPONENT sa = Source_Maxwell_3( /* flux in n/s/cm2/st/AA */
                                        Lmin=1, Lmax=11, dist=1, focus_xw=0.1, focus_yh=0.1, xwidth=0.01, yheight=0.01,
                                        T1=216.8,I1=1.24e+13,T2=33.9,I2=1.02e+13, T3=16.7 ,I3=3.0423e+12)
        AT (0,0,0) ABSOLUTE
        EXTEND %{
            t=1e-3*rand01();
            flag=(double)mcget_run_num();
        %}
        COMPONENT split_point = Arm() AT(0,0,0) RELATIVE PREVIOUS
        COMPONENT m1 = Monitor_nD(filename="o1", xwidth=0.2, yheight=0.2,
                                  options="lambda limits=[1 11] bins=100 parallel", bins=40)
        AT (0,0,0) ABSOLUTE
        COMPONENT m2 = Monitor_nD(filename="o2", xwidth=0.2, yheight=0.2, options="x y, parallel", bins=40)
        AT (0,0,0) ABSOLUTE
        COMPONENT m3 = Monitor_nD(filename="o3", xwidth=0.2, yheight=0.2, options="t limits=[0 1e-3]parallel", bins=40)
        AT (0,0,0) ABSOLUTE
        COMPONENT m4 = Monitor_nD(filename="o4", xwidth=0.2, yheight=0.2, options="E limits=[0 82] parallel", bins=40)
        AT (0,0,0) ABSOLUTE
        END
        """
        instr = parse_instr_string(instr_source)
        self.assertEqual(instr.name, 'Test_MCPL_output')

        flag_params = tuple(ComponentParameter(k, Expr.str(v)) for k, v in [
            ('userflag', '"flag"'), ('userflagcomment', '"Neutron Id"')])
        before, after = instr.mcpl_split('split_point', filename='test_mcpl_split', output_parameters=flag_params)
        self.assertEqual(before.name, 'Test_MCPL_output_first')
        self.assertEqual(after.name, 'Test_MCPL_output_second')
        self.assertTrue(isinstance(before, Instr))
        self.assertTrue(isinstance(after, Instr))
        self.assertEqual(len(before.components), 3)
        self.assertEqual(len(after.components), 5)
        for i, name in enumerate(('Origin', 'sa', 'split_point')):
            self.assertEqual(before.components[i].name, name)
        for i, name in enumerate(('split_point', 'm1', 'm2', 'm3', 'm4')):
            self.assertEqual(after.components[i].name, name)

        # If we compile and run the parsed instrument, or the split instrument parts, we should get the same results
        target = CBinaryTarget(mpi=False, acc=False, count=1, nexus=False)
        config = dict(default_main=True, enable_trace=False, portable=False, include_runtime=True,
                      embed_instrument_file=False, verbose=False)
        with TemporaryDirectory() as directory:
            expected_binaries = [Path(directory).joinpath(f'{x.name}.out') for x in (instr, before, after)]
            for obj in (instr, before, after):
                try:
                    compile_instrument(obj, target, directory, generator=MCSTAS_GENERATOR, config=config)
                except RuntimeError as e:
                    log.error(f'Failed to compile instrument: {e}')
                    raise e

            for binary in expected_binaries:
                self.assertTrue(binary.exists())
                self.assertTrue(binary.is_file())
                self.assertTrue(access(binary, R_OK))

            # Run the instrument and check that the output is the same
            seed = randint(1000, 2**32 - 1)
            common = f'--seed {seed} -n 10000'
            run_compiled_instrument(expected_binaries[0], target, f"--dir {directory}/instr {common}")
            mcpl_file = Path(directory).joinpath('instr.mcpl')
            run_compiled_instrument(expected_binaries[1], target, f"--dir {directory}/before {common} mcpl_filename={mcpl_file}")
            # Depending on how MCPL was built, it might have gzipped the output file
            if not mcpl_file.exists():
                mcpl_file = Path(directory).joinpath('instr.mcpl.gz')
            self.assertTrue(mcpl_file.exists())
            run_compiled_instrument(expected_binaries[2], target,
                                    f"--dir {directory}/after {common} mcpl_filename={mcpl_file}")

            # Now check that the instr and after outputs are the same
            instr_files = list(Path(directory).joinpath('instr').glob('o*'))
            after_files = list(Path(directory).joinpath('after').glob('o*'))
            self.assertEqual(len(instr_files), len(after_files))
            for file in instr_files:
                self.assertEqual(len([x for x in after_files if x.name == file.name]), 1)
            for file in after_files:
                self.assertEqual(len([x for x in instr_files if x.name == file.name]), 1)
            instr_files = [[x for x in instr_files if x.name == file.name][0] for file in after_files]
            for instr_file, after_file in zip(instr_files, after_files):
                # The files must be the same except for any header information:
                instr_data = read_mccode_dat(instr_file)
                after_data = read_mccode_dat(after_file)
                self.assertTrue(allclose(instr_data.data, after_data.data))



import confuse
# Try and simplify handling configuration values need for, e.g., compiling different versions of the runtimes
# under different operating systems, while allowing a user to 'easily' override defaults if necessary.
#
# Any platform independent configuration settings can go in 'config_default.yaml'
config = confuse.LazyConfig('mccode_antlr', __name__)

# use environment variables specified as 'MCCODE_XYZ' as configuration entries 'xyz'
config.set_env()


# Add platform specific values from the local config_platforms.yaml
def _platform_defaults():
    import platform
    import yaml
    from importlib.resources import files, as_file

    platforms_file = files(__name__).joinpath('platforms.yaml')
    if not platforms_file.is_file():
        raise RuntimeError(f"Can not locate platforms.yaml in module files (looking for {platforms_file}")

    with as_file(platforms_file) as file:
        with open(file, 'r') as data:
            platform_configs = yaml.safe_load(data)

    system = platform.system()  # "Linux", "Darwin", "Windows", ...
    if system not in platform_configs:
        raise RuntimeError(f"Unknown system '{system}' for default configuration")

    return platform_configs[system]


# By using the 'add' method, we set these as the *lowest* priority. Any user/system files will override:
config.add(_platform_defaults())



from ..grammar import McInstrParser, McInstrVisitor
from ..common import InstrumentParameter, MetaData, Expr
from .instr import Instr
from .instance import Instance
from .jump import Jump
from zenlog import log


def literal_string(ctx):
    start_token, stop_token = ctx.start, ctx.stop
    stream = start_token.getInputStream()
    return stream.getText(start_token.start, stop_token.stop)


class InstrVisitor(McInstrVisitor):
    def __init__(self, parent, filename, destination=None, allow_assignment=False):
        self.parent = parent
        self.filename = filename
        self.state = Instr()
        self.current_comp = None
        self.current_instance_name = None
        self.destination = destination

    def visitProg(self, ctx: McInstrParser.ProgContext):
        self.state = Instr()
        self.visit(ctx.instrument_definition())
        return self.state

    def visitInstrument_definition(self, ctx: McInstrParser.Instrument_definitionContext):
        self.state.name = str(ctx.Identifier())
        self.visitChildren(ctx)
        self.state.determine_groups()

    def visitInstrument_parameters(self, ctx: McInstrParser.Instrument_parametersContext):
        for param in ctx.params:
            self.state.add_parameter(self.visit(param))

    def getInstrument_parameter(self, ctx: McInstrParser.Instrument_parameterContext):
        return self.visit(ctx)

    def visitInstrument_metadata(self, ctx: McInstrParser.Instrument_metadataContext):
        for metadata_context in ctx.metadata():
            mime, name, metadata = self.visit(metadata_context)
            metadata = MetaData.from_instrument_tokens(source=self.state.name, mimetype=mime, name=name, value=metadata)
            self.state.add_metadata(metadata)

    def visitInstrumentParameterDouble(self, ctx: McInstrParser.InstrumentParameterDoubleContext):
        name = str(ctx.Identifier())
        unit = None if ctx.instrument_parameter_unit() is None else self.visit(ctx.instrument_parameter_unit())
        value = None if ctx.Assign() is None else self.visit(ctx.expr())
        return InstrumentParameter(name, unit, Expr.float(value))

    def visitInstrumentParameterInteger(self, ctx: McInstrParser.InstrumentParameterIntegerContext):
        name = str(ctx.Identifier())
        unit = None if ctx.instrument_parameter_unit() is None else self.visit(ctx.instrument_parameter_unit())
        value = None if ctx.Assign() is None else self.visit(ctx.expr())
        return InstrumentParameter(name, unit, Expr.int(value))

    def visitInstrumentParameterString(self, ctx: McInstrParser.InstrumentParameterStringContext):
        name = str(ctx.Identifier())
        unit = None if ctx.instrument_parameter_unit() is None else self.visit(ctx.instrument_parameter_unit())
        value = None
        if ctx.Assign() is not None:
            value = 'NULL' if ctx.StringLiteral() is None else str(ctx.StringLiteral())
        return InstrumentParameter(name, unit, Expr.str(value))

    def visitInstrument_parameter_unit(self, ctx: McInstrParser.Instrument_parameter_unitContext):
        return str(ctx.StringLiteral())

    def visitInstrument_trace(self, ctx: McInstrParser.Instrument_traceContext):
        return self.visitChildren(ctx)

    def visitInstrument_trace_include(self, ctx: McInstrParser.Instrument_trace_includeContext):
        quoted_filename = str(ctx.StringLiteral())
        if self.destination is not None:
            log.critical(f'including {quoted_filename} from {self.filename}, which is itself included from {self.destination.name}')
            log.critical('Expect component referencing errors, as the implementation does not cover this use case.')
        instr = self.parent.get_instrument(quoted_filename.strip('"'), destination=self.state)
        # TODO work out how/what to copy from the other instrument into this one
        self.state.add_included(instr.name)
        for par in instr.parameters:
            self.state.add_parameter(par, ignore_repeated=True)
        for meta in instr.metadata:
            self.state.add_metadata(meta)
        if len(instr.declare):
            self.state.declare += instr.declare
        if len(instr.user):
            self.state.user += instr.user
        if len(instr.initialize):
            self.state.initialize += instr.initialize
        if len(instr.save):
            self.state.save += instr.save
        if len(instr.final):
            self.state.final += instr.final
        # McCode3 parsed everything in one memory space, so used some trickery to include
        # component instances from one instrument into another. Here we can be a bit more straightforward
        for instance in instr.components:
            if not instance.removable:
                self.state.add_component(instance)
        # Group membership is determined after all parsing, so nothing to do here

    def visitComponent_instance(self, ctx: McInstrParser.Component_instanceContext):
        from ..comp import Comp
        from ..instr.orientation import Angles
        name = self.visit(ctx.instance_name())
        self.current_instance_name = name
        comp = self.visit(ctx.component_type())
        if not isinstance(comp, (Comp, Instance)):
            raise RuntimeError(f'Undefined component type {type(comp)}')
        is_ref = isinstance(comp, Instance)
        self.current_comp = comp.type if is_ref else comp
        at = self.visit(ctx.place())
        if ctx.orientation() is not None:
            rotate = self.visit(ctx.orientation())
        else:
            # In the case of "AT (x, y, z) ABSOLUTE" or "AT (x, y, z) RELATIVE identifier"
            # We must use *the same* relative information for the rotation -- at[1] is None or a valid instance:
            rotate = (Angles(Expr.int(0), Expr.int(0), Expr.int(0)), at[1])
        # Construct a new instance, possibly copying values from an existing instance:
        instance = Instance.from_instance(name, comp, at, rotate) if is_ref else Instance(name, comp, at, rotate)
        if ctx.instance_parameters() is not None:
            for param_name, param_value in self.visit(ctx.instance_parameters()):
                instance.set_parameter(param_name, param_value, overwrite=is_ref)
        if ctx.Removable() is not None:
            instance.REMOVABLE()
        if ctx.Cpu() is not None:
            instance.CPU()
        if ctx.split() is not None:
            instance.SPLIT(self.visit(ctx.split()))
        if ctx.when() is not None:
            instance.WHEN(self.visit(ctx.when()))
        if ctx.groupref() is not None:
            instance.GROUP(self.visit(ctx.groupref()))
        if ctx.extend() is not None:
            instance.EXTEND(self.visit(ctx.extend()))
        if ctx.jumps() is not None:
            instance.JUMP(*self.visit(ctx.jumps()))
        if ctx.metadata() is not None:
            # deal with definition vs instance metadata here?
            for metadata_context in ctx.metadata():
                mime, name, metadata = self.visit(metadata_context)
                metadata = MetaData.from_instance_tokens(source=instance.name, mimetype=mime, name=name, value=metadata)
                instance.add_metadata(metadata)
        # Include this instantiated component instance in the instrument components list
        if self.destination is None or not instance.removable:
            # if this _is_ an included instrument, any REMOVABLE component instances should not be added
            # TODO we don't need to populate the whole Instance object if this branch is not selected
            self.state.add_component(instance)
        self.current_comp = None
        self.current_instance_name = None

    def visitInstanceNameCopyIdentifier(self, ctx: McInstrParser.InstanceNameCopyIdentifierContext):
        return f'{ctx.Identifier()}_{len(self.state.components)+1}'

    def visitInstanceNameCopy(self, ctx: McInstrParser.InstanceNameCopyContext):
        return f'Comp_{len(self.state.components) + 1}'

    def visitInstanceNameIdentifier(self, ctx: McInstrParser.InstanceNameIdentifierContext):
        return str(ctx.Identifier())

    def visitComponentTypeCopy(self, ctx: McInstrParser.ComponentTypeCopyContext):
        return self.visit(ctx.component_ref())

    def visitComponentTypeIdentifier(self, ctx: McInstrParser.ComponentTypeIdentifierContext):
        return self.parent.get_component(str(ctx.Identifier()), current_instance_name=self.current_instance_name)

    def visitInstance_parameters(self, ctx: McInstrParser.Instance_parametersContext):
        return [self.visit(p) for p in ctx.params]

    def visitInstanceParameterExpr(self, ctx: McInstrParser.InstanceParameterExprContext):
        from ..common import DataType
        name = str(ctx.Identifier())
        value = self.visit(ctx.expr())
        default = self.current_comp.get_parameter(name)
        if default is None:
            raise RuntimeError(f'{name} is not a known DEFINITION or SETTING parameter for {self.current_comp.name}')
        if DataType.undefined == value.data_type:
            value.data_type = default.value.data_type
            value.shape_type = default.value.shape_type
        return name, value

    def visitInstanceParameterNull(self, ctx: McInstrParser.InstanceParameterNullContext):
        from ..common import DataType
        name = str(ctx.Identifier())
        value = Expr.str('NULL')
        default = self.current_comp.get_parameter(name)
        if default is None:
            raise RuntimeError(f'{name} is not a known DEFINITION or SETTING parameter for {self.current_comp.name}')
        if DataType.undefined == value.data_type:
            value.data_type = default.value.data_type
        return name, value

    def visitInstanceParameterVector(self, ctx: McInstrParser.InstanceParameterVectorContext):
        from ..common import DataType
        name = str(ctx.Identifier())
        value = self.visit(ctx.initializerlist())
        value.data_type = DataType.float
        return name, value

    def visitSplit(self, ctx: McInstrParser.SplitContext):
        return Expr.int(10) if ctx.expr() is None else self.visit(ctx.expr())

    def visitWhen(self, ctx: McInstrParser.WhenContext):
        return self.visit(ctx.expr())

    def visitPlace(self, ctx: McInstrParser.PlaceContext):
        from mccode_antlr.instr.orientation import Vector
        vector = self.visit(ctx.coords())
        relative = self.visit(ctx.reference())
        return Vector(*vector), relative

    def visitOrientation(self, ctx: McInstrParser.OrientationContext):
        from mccode_antlr.instr.orientation import Angles
        angles = self.visit(ctx.coords())
        relative = self.visit(ctx.reference())
        return Angles(*angles), relative

    def visitGroupref(self, ctx: McInstrParser.GrouprefContext):
        return str(ctx.Identifier())

    def visitJumps(self, ctx: McInstrParser.JumpsContext):
        return [self.visit(j) for j in ctx.jump()]

    def visitJump(self, ctx: McInstrParser.JumpContext):
        name, index = self.visit(ctx.jumpname())
        iterate = ctx.Iterate() is not None  # and ctx.When() is None
        condition = self.visit(ctx.expr())
        return Jump(name, index, iterate, condition)

    def visitJumpPrevious(self, ctx: McInstrParser.JumpPreviousContext):
        i = ctx.IntegerLiteral()
        return ("PREVIOUS", -1) if i is None else (f"PREVIOUS_{i}", -int(i))

    def visitJumpMyself(self, ctx: McInstrParser.JumpMyselfContext):
        return "MYSELF", 0

    def visitJumpNext(self, ctx: McInstrParser.JumpNextContext):
        i = ctx.IntegerLiteral()
        return ("NEXT", 1) if i is None else (f"NEXT_{i}", int(i))

    def visitJumpIdentifier(self, ctx: McInstrParser.JumpIdentifierContext):
        return str(ctx.Identifier()), 0

    def visitComponent_ref(self, ctx: McInstrParser.Component_refContext):
        if ctx.Previous() is not None:
            count = 1 if ctx.IntegerLiteral() is None else int(str(ctx.IntegerLiteral()))
            # Any included component can be referred to -- REMOVABLE components in an included instrument
            # were _not_ included into the state. Include REMOVABLE components _are_ in the state.
            instances = len(self.state.components)
            if count <= instances:
                return self.state.last_component(count, removable_ok=True)
            elif self.destination is not None:
                return self.destination.last_component(count - instances, removable_ok=True)
            else:
                log.error(f'Too large PREVIOUS count {count} for instrument with {instances} component instances')
        name = str(ctx.Identifier())
        if any(inst.name == name for inst in self.state.components):
            return self.state.get_component(name)
        elif self.destination is not None:
            return self.destination.get_component(name)
        else:
            log.error(f'Unknown component reference for instance named {name}')

    def visitCoords(self, ctx: McInstrParser.CoordsContext):
        # FIXME 2023-10-16 previously Cooridnate parsing forced all returned Expression objects to be floats
        #   while this is 'correct', it broke the ability to use instrument parameters in component placement
        #   and orientation expressions. I think expression parsing is now implemented correctly for all variants
        #   So _hopefully_ the explicit cast-to-float is not required any longer.
        # A coordinate is _always_ a float, even when represented by an expression or identifier
        # Actually implement the visitExpr variants correctly?
        # return tuple([Expr.float(self.visit(x)) for x in ctx.expr()])
        return tuple([self.visit(x) for x in ctx.expr()])

    def visitReference(self, ctx: McInstrParser.ReferenceContext):
        # ABSOLUTE or RELATIVE ABSOLUTE -> None
        return self.visit(ctx.component_ref()) if ctx.Absolute() is None else None

    def visitDependency(self, ctx: McInstrParser.DependencyContext):
        # store the flag without its surrounding quotes
        self.state.DEPENDENCY(str(ctx.StringLiteral())[1:-1])

    def visitDeclareBlock(self, ctx: McInstrParser.DeclareBlockContext):
        self.state.DECLARE(self.visit(ctx.unparsed_block()))

    def visitUservars(self, ctx: McInstrParser.UservarsContext):
        self.state.USERVARS(self.visit(ctx.unparsed_block()))

    def visitInitializeBlock(self, ctx: McInstrParser.InitializeBlockContext):
        self.state.INITIALIZE(self.visit(ctx.unparsed_block()))

    def visitSaveBlock(self, ctx: McInstrParser.SaveBlockContext):
        self.state.SAVE(self.visit(ctx.unparsed_block()))

    def visitFinallyBlock(self, ctx: McInstrParser.FinallyBlockContext):
        self.state.FINALLY(self.visit(ctx.unparsed_block()))

    def visitDeclareBlockCopy(self, ctx: McInstrParser.DeclareBlockCopyContext):
        line_number = None if ctx.start is None else ctx.start.line
        raise RuntimeError(f"{self.filename}: {line_number} -- DECLARE COPY only valid in comp files")
        # copy_from = self.parent.get_instrument(str(ctx.Identifier()))
        # self.state.DECLARE(copy_from.declare, self.visit(ctx.unparsed_block()))

    def visitInitializeBlockCopy(self, ctx: McInstrParser.InitializeBlockCopyContext):
        line_number = None if ctx.start is None else ctx.start.line
        raise RuntimeError(f"{self.filename}: {line_number} -- INITIALIZE COPY only valid in comp files")
        # copy_from = self.parent.get_instrument(str(ctx.Identifier()))
        # self.state.INITIALIZE(copy_from.initialize, self.visit(ctx.unparsed_block()))

    def visitSaveBlockCopy(self, ctx: McInstrParser.SaveBlockCopyContext):
        line_number = None if ctx.start is None else ctx.start.line
        raise RuntimeError(f"{self.filename}: {line_number} -- SAVE COPY only valid in comp files")
        # copy_from = self.parent.get_instrument(str(ctx.Identifier()))
        # self.state.SAVE(copy_from.save, self.visit(ctx.unparsed_block()))

    def visitFinallyBlockCopy(self, ctx: McInstrParser.FinallyBlockCopyContext):
        line_number = None if ctx.start is None else ctx.start.line
        raise RuntimeError(f"{self.filename}: {line_number} -- FINALLY COPY only valid in comp files")
        # copy_from = self.parent.get_instrument(str(ctx.Identifier()))
        # self.state.FINALLY(copy_from.final, self.visit(ctx.unparsed_block()))

    def visitExtend(self, ctx: McInstrParser.ExtendContext):
        return self.visit(ctx.unparsed_block())

    def visitMetadata(self, ctx: McInstrParser.MetadataContext):
        filename, line_number, metadata = self.visit(ctx.unparsed_block())
        # ctx.mime and ctx.name are _either_ identifiers (no double quotes) or string literals (double quotes)
        # so we need to strip the quotes from the string literals, but not the identifiers
        mime = ctx.mime.text if ctx.mime.type == McInstrParser.Identifier else ctx.mime.text[1:-1]
        name = ctx.name.text if ctx.name.type == McInstrParser.Identifier else ctx.name.text[1:-1]
        return mime, name, metadata

    def visitUnparsed_block(self, ctx: McInstrParser.Unparsed_blockContext):
        # We want to extract the source-file line number (and filename) for use in the C-preprocessor
        # via `#line {number} "{filename}"` directives, for more expressive error handling
        line_number = None if ctx.start is None else ctx.start.line
        if line_number is None:
            print(f'Why is line none for {self.filename} {ctx.UnparsedBlock()}')
        return self.filename, line_number,  str(ctx.UnparsedBlock())[2:-2]

    def visitShell(self, ctx: McInstrParser.ShellContext):
        from subprocess import run
        args = str(ctx.StringLiteral()).split(' ')
        run(args, shell=True, check=True)

    def visitSearchPath(self, ctx: McInstrParser.SearchPathContext):
        self.parent.handle_search_keyword(str(ctx.StringLiteral()))

    def visitSearchShell(self, ctx: McInstrParser.SearchShellContext):
        from subprocess import run
        args = str(ctx.StringLiteral()).split()
        res = run(args, shell=True, capture_output=True, check=True)
        for specs in res.stdout.decode().split('\n'):
            self.parent.handle_search_keyword(specs)

    # TODO Make this and the identical list of visitors in comp/visitor.py a single definition ... somehow
    # FIXME There *are* no statements in McCode, so all identifiers always produce un-parsable values.
    def visitAssignment(self, ctx: McInstrParser.AssignmentContext):
        line_number = None if ctx.start is None else ctx.start.line
        raise RuntimeError(f"{self.filename}: {line_number} -- assignment statements are not (yet) supported")

    def getExpr(self, ctx: McInstrParser.ExprContext):
        return self.visit(ctx)

    # TODO (maybe) Add control and statements into McCode, requiring some form of global stack.
    def visitExpressionUnaryPM(self, ctx: McInstrParser.ExpressionUnaryPMContext):
        right = self.visit(ctx.expr())
        if isinstance(right, str):
            return '-' + right if ctx.Plus() is None else right
        return -right if ctx.Plus() is None else right

    def visitExpressionGrouping(self, ctx: McInstrParser.ExpressionGroupingContext):
        from ..common import UnaryOp
        return Expr(UnaryOp('__group__', self.visit(ctx.expr())))

    def visitExpressionFloat(self, ctx: McInstrParser.ExpressionFloatContext):
        return Expr.float(str(ctx.FloatingLiteral()))

    def visitExpressionPointerAccess(self, ctx: McInstrParser.ExpressionPointerAccessContext):
        from ..common import BinaryOp, Value, ObjectType
        pointer = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier))
        return Expr(BinaryOp('__pointer_access__', pointer, self.visit(ctx.expr())))

    def visitExpressionStructAccess(self, ctx: McInstrParser.ExpressionStructAccessContext):
        from ..common import BinaryOp, Value, ObjectType
        struct = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier))
        return Expr(BinaryOp('__struct_access__', struct, self.visit(ctx.expr())))

    def visitExpressionArrayAccess(self, ctx: McInstrParser.ExpressionArrayAccessContext):
        from ..common import BinaryOp, Value, ShapeType, ObjectType
        array = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier, shape_type=ShapeType.vector))
        return Expr(BinaryOp('__getitem__', array, self.visit(ctx.expr())))

    def visitExpressionIdentifier(self, ctx: McInstrParser.ExpressionIdentifierContext):
        from ..common import Value, DataType, ObjectType, parameter_name_present
        # check if this identifier is an InstrumentParameter name:
        name = str(ctx.Identifier())
        inst_par = self.state.get_parameter(name, None)
        obj = ObjectType.parameter if inst_par is not None else ObjectType.identifier
        dat = inst_par.value.data_type if inst_par is not None else DataType.undefined
        return Expr(Value(name, data_type=dat, object_type=obj))

    def visitExpressionInteger(self, ctx: McInstrParser.ExpressionIntegerContext):
        return Expr.int(str(ctx.IntegerLiteral()))

    def visitExpressionZero(self, ctx: McInstrParser.ExpressionZeroContext):
        return Expr.int(0)

    def visitExpressionExponentiation(self, ctx: McInstrParser.ExpressionExponentiationContext):
        base = self.visit(ctx.base)
        exponent = self.visit(ctx.exponent)
        return base ** exponent

    def visitExpressionBinaryPM(self, ctx: McInstrParser.ExpressionBinaryPMContext):
        left = self.visit(ctx.left)
        right = self.visit(ctx.right)
        return left + right if ctx.Minus() is None else left - right

    def visitExpressionFunctionCall(self, ctx: McInstrParser.ExpressionFunctionCallContext):
        from ..common import BinaryOp, Value, ObjectType
        function = Value(str(ctx.Identifier()), object_type=ObjectType.function)
        args = [self.visit(arg).expr[0] for arg in ctx.args]  # each is a Value, UnaryOp, or BinaryOp?
        return Expr(BinaryOp('__call__', function, args))

    def visitExpressionBinaryMD(self, ctx: McInstrParser.ExpressionBinaryMDContext):
        left, right = self.visit(ctx.left), self.visit(ctx.right)
        return left * right if ctx.Div() is None else left / right

    def visitInitializerlist(self, ctx: McInstrParser.InitializerlistContext):
        from ..common import Value, ObjectType, ShapeType
        values = [self.visit(x).expr[0].value for x in ctx.values]
        return Expr(Value(values, object_type=ObjectType.initializer_list, shape_type=ShapeType.vector))

    def visitExpressionUnaryLogic(self, ctx: McInstrParser.ExpressionUnaryLogicContext):
        from ..common import UnaryOp
        expr = self.visit(ctx.expr())
        op = 'unknown'
        if ctx.Not() is not None:
            op = '__not__'
        return Expr(UnaryOp(op, expr))

    def visitExpressionBinaryLogic(self, ctx: McInstrParser.ExpressionBinaryLogicContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        op = 'unknown'
        if ctx.AndAnd() is not None:
            op = '__and__'
        elif ctx.OrOr() is not None:
            op = '__or__'
        return Expr(BinaryOp(op, left, right))

    def visitExpressionTrinaryLogic(self, ctx: McInstrParser.ExpressionTrinaryLogicContext):
        from ..common import TrinaryOp
        test, true, false = [self.visit(x) for x in (ctx.test, ctx.true, ctx.false)]
        return Expr(TrinaryOp('__trinary__', test, true, false))

    def visitExpressionBinaryEqual(self, ctx: McInstrParser.ExpressionBinaryEqualContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__eq__', left, right))

    def visitExpressionBinaryLessEqual(self, ctx: McInstrParser.ExpressionBinaryLessEqualContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__le__', left, right))

    def visitExpressionBinaryGreaterEqual(self, ctx: McInstrParser.ExpressionBinaryGreaterEqualContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__ge__', left, right))

    def visitExpressionBinaryLess(self, ctx: McInstrParser.ExpressionBinaryLessContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__lt__', left, right))

    def visitExpressionBinaryGreater(self, ctx: McInstrParser.ExpressionBinaryGreaterContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__gt__', left, right))

    def visitExpressionString(self, ctx: McInstrParser.ExpressionStringContext):
        strings = ''.join(str(sl).strip('"') for sl in ctx.StringLiteral())
        return Expr.str(f'"{strings}"')

    def visitExpressionPrevious(self, ctx: McInstrParser.ExpressionPreviousContext):
        # The very-special no-good expression use of PREVIOUS where it is replaced by the last component's name
        if len(self.state.components):
            return Expr.str(self.state.components[-1].name)
        elif self.destination is not None and len(self.destination.components):
            return Expr.str(self.destination.components[-1].name)
        raise RuntimeError('PREVIOUS keyword used in expression before any components defined')

    def visitExpressionMyself(self, ctx: McInstrParser.ExpressionMyselfContext):
        # The even-worse expression use of MYSELF to refer to the current being-constructed component's name
        return Expr.str(self.current_instance.name)


class InstrParametersVisitor(McInstrVisitor):
    """A visitor which takes a full parse tree and extracts only the instrument parameters"""
    def __init__(self):
        self.state = Instr()

    def visitProg(self, ctx: McInstrParser.ProgContext):
        self.state = Instr()
        self.visit(ctx.instrument_definition())
        return self.state.parameters

    def visitInstrument_definition(self, ctx: McInstrParser.Instrument_definitionContext):
        self.state.name = str(ctx.Identifier())
        self.visitChildren(ctx)

    def visitInstrument_parameters(self, ctx: McInstrParser.Instrument_parametersContext):
        for param in ctx.params:
            self.state.add_parameter(self.visit(param))

    def visitInstrumentParameterDouble(self, ctx: McInstrParser.InstrumentParameterDoubleContext):
        name = str(ctx.Identifier())
        unit = None if ctx.instrument_parameter_unit() is None else self.visit(ctx.instrument_parameter_unit())
        value = None if ctx.Assign() is None else self.visit(ctx.expr())
        return InstrumentParameter(name, unit, Expr.float(value))

    def visitInstrumentParameterInteger(self, ctx: McInstrParser.InstrumentParameterIntegerContext):
        name = str(ctx.Identifier())
        unit = None if ctx.instrument_parameter_unit() is None else self.visit(ctx.instrument_parameter_unit())
        value = None if ctx.Assign() is None else self.visit(ctx.expr())
        return InstrumentParameter(name, unit, Expr.int(value))

    def visitInstrumentParameterString(self, ctx: McInstrParser.InstrumentParameterStringContext):
        name = str(ctx.Identifier())
        unit = None if ctx.instrument_parameter_unit() is None else self.visit(ctx.instrument_parameter_unit())
        value = None
        if ctx.Assign() is not None:
            value = 'NULL' if ctx.StringLiteral() is None else str(ctx.StringLiteral())
        return InstrumentParameter(name, unit, Expr.str(value))

    def visitInstrument_parameter_unit(self, ctx: McInstrParser.Instrument_parameter_unitContext):
        return str(ctx.StringLiteral())

    def visitExpressionUnaryPM(self, ctx: McInstrParser.ExpressionUnaryPMContext):
        right = self.visit(ctx.expr())
        if isinstance(right, str):
            return '-' + right if ctx.Plus() is None else right
        return -right if ctx.Plus() is None else right

    def visitExpressionGrouping(self, ctx: McInstrParser.ExpressionGroupingContext):
        from ..common import UnaryOp
        return Expr(UnaryOp('__group__', self.visit(ctx.expr())))

    def visitExpressionFloat(self, ctx: McInstrParser.ExpressionFloatContext):
        return Expr.float(str(ctx.FloatingLiteral()))

    def visitExpressionPointerAccess(self, ctx: McInstrParser.ExpressionPointerAccessContext):
        from ..common import BinaryOp, Value, ObjectType
        pointer = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier))
        return Expr(BinaryOp('__pointer_access__', pointer, self.visit(ctx.expr())))

    def visitExpressionStructAccess(self, ctx: McInstrParser.ExpressionStructAccessContext):
        from ..common import BinaryOp, Value, ObjectType
        struct = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier))
        return Expr(BinaryOp('__struct_access__', struct, self.visit(ctx.expr())))

    def visitExpressionArrayAccess(self, ctx: McInstrParser.ExpressionArrayAccessContext):
        from ..common import BinaryOp, Value, ShapeType, ObjectType
        array = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier, shape_type=ShapeType.vector))
        return Expr(BinaryOp('__getitem__', array, self.visit(ctx.expr())))

    def visitExpressionIdentifier(self, ctx: McInstrParser.ExpressionIdentifierContext):
        from ..common import Value, DataType, ObjectType, parameter_name_present
        # check if this identifier is an InstrumentParameter name:
        name = str(ctx.Identifier())
        inst_par = self.state.get_parameter(name, None)
        obj = ObjectType.parameter if inst_par is not None else ObjectType.identifier
        dat = inst_par.value.data_type if inst_par is not None else DataType.undefined
        return Expr(Value(name, data_type=dat, object_type=obj))

    def visitExpressionInteger(self, ctx: McInstrParser.ExpressionIntegerContext):
        return Expr.int(str(ctx.IntegerLiteral()))

    def visitExpressionZero(self, ctx: McInstrParser.ExpressionZeroContext):
        return Expr.int(0)

    def visitExpressionExponentiation(self, ctx: McInstrParser.ExpressionExponentiationContext):
        base = self.visit(ctx.base)
        exponent = self.visit(ctx.exponent)
        return base ** exponent

    def visitExpressionBinaryPM(self, ctx: McInstrParser.ExpressionBinaryPMContext):
        left = self.visit(ctx.left)
        right = self.visit(ctx.right)
        return left + right if ctx.Minus() is None else left - right

    def visitExpressionFunctionCall(self, ctx: McInstrParser.ExpressionFunctionCallContext):
        from ..common import BinaryOp, Value, ObjectType
        function = Value(str(ctx.Identifier()), object_type=ObjectType.function)
        args = [self.visit(arg).expr[0] for arg in ctx.args]  # each is a Value, UnaryOp, or BinaryOp?
        return Expr(BinaryOp('__call__', function, args))

    def visitExpressionBinaryMD(self, ctx: McInstrParser.ExpressionBinaryMDContext):
        left, right = self.visit(ctx.left), self.visit(ctx.right)
        return left * right if ctx.Div() is None else left / right

    def visitInitializerlist(self, ctx: McInstrParser.InitializerlistContext):
        from ..common import Value, ObjectType, ShapeType
        values = [self.visit(x).expr[0].value for x in ctx.values]
        return Expr(Value(values, object_type=ObjectType.initializer_list, shape_type=ShapeType.vector))

    def visitExpressionUnaryLogic(self, ctx: McInstrParser.ExpressionUnaryLogicContext):
        from ..common import UnaryOp
        expr = self.visit(ctx.expr())
        op = 'unknown'
        if ctx.Not() is not None:
            op = '__not__'
        return Expr(UnaryOp(op, expr))

    def visitExpressionBinaryLogic(self, ctx: McInstrParser.ExpressionBinaryLogicContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        op = 'unknown'
        if ctx.AndAnd() is not None:
            op = '__and__'
        elif ctx.OrOr() is not None:
            op = '__or__'
        return Expr(BinaryOp(op, left, right))

    def visitExpressionTrinaryLogic(self, ctx: McInstrParser.ExpressionTrinaryLogicContext):
        from ..common import TrinaryOp
        test, true, false = [self.visit(x) for x in (ctx.test, ctx.true, ctx.false)]
        return Expr(TrinaryOp('__trinary__', test, true, false))

    def visitExpressionBinaryEqual(self, ctx: McInstrParser.ExpressionBinaryEqualContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__eq__', left, right))

    def visitExpressionBinaryLessEqual(self, ctx: McInstrParser.ExpressionBinaryLessEqualContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__le__', left, right))

    def visitExpressionBinaryGreaterEqual(self, ctx: McInstrParser.ExpressionBinaryGreaterEqualContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__ge__', left, right))

    def visitExpressionBinaryLess(self, ctx: McInstrParser.ExpressionBinaryLessContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__lt__', left, right))

    def visitExpressionBinaryGreater(self, ctx: McInstrParser.ExpressionBinaryGreaterContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__gt__', left, right))

    def visitExpressionString(self, ctx: McInstrParser.ExpressionStringContext):
        strings = ''.join(str(sl).strip('"') for sl in ctx.StringLiteral())
        return Expr.str(f'"{strings}"')

    def visitExpressionPrevious(self, ctx: McInstrParser.ExpressionPreviousContext):
        # The very-special no-good expression use of PREVIOUS where it is replaced by the last component's name
        if len(self.state.components):
            return Expr.str(self.state.components[-1].name)
        elif self.destination is not None and len(self.destination.components):
            return Expr.str(self.destination.components[-1].name)
        raise RuntimeError('PREVIOUS keyword used in expression before any components defined')

    def visitExpressionMyself(self, ctx: McInstrParser.ExpressionMyselfContext):
        # The even-worse expression use of MYSELF to refer to the current being-constructed component's name
        return Expr.str(self.current_instance.name)


# Generated from /home/g/Code/mccode_antlr-antlr/mccode_antlr/grammar/McComp.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .McCompParser import McCompParser
else:
    from McCompParser import McCompParser

# This class defines a complete listener for a parse tree produced by McCompParser.
class McCompListener(ParseTreeListener):

    # Enter a parse tree produced by McCompParser#prog.
    def enterProg(self, ctx:McCompParser.ProgContext):
        pass

    # Exit a parse tree produced by McCompParser#prog.
    def exitProg(self, ctx:McCompParser.ProgContext):
        pass


    # Enter a parse tree produced by McCompParser#ComponentDefineNew.
    def enterComponentDefineNew(self, ctx:McCompParser.ComponentDefineNewContext):
        pass

    # Exit a parse tree produced by McCompParser#ComponentDefineNew.
    def exitComponentDefineNew(self, ctx:McCompParser.ComponentDefineNewContext):
        pass


    # Enter a parse tree produced by McCompParser#ComponentDefineCopy.
    def enterComponentDefineCopy(self, ctx:McCompParser.ComponentDefineCopyContext):
        pass

    # Exit a parse tree produced by McCompParser#ComponentDefineCopy.
    def exitComponentDefineCopy(self, ctx:McCompParser.ComponentDefineCopyContext):
        pass


    # Enter a parse tree produced by McCompParser#TraceBlock.
    def enterTraceBlock(self, ctx:McCompParser.TraceBlockContext):
        pass

    # Exit a parse tree produced by McCompParser#TraceBlock.
    def exitTraceBlock(self, ctx:McCompParser.TraceBlockContext):
        pass


    # Enter a parse tree produced by McCompParser#TraceBlockCopy.
    def enterTraceBlockCopy(self, ctx:McCompParser.TraceBlockCopyContext):
        pass

    # Exit a parse tree produced by McCompParser#TraceBlockCopy.
    def exitTraceBlockCopy(self, ctx:McCompParser.TraceBlockCopyContext):
        pass


    # Enter a parse tree produced by McCompParser#component_parameter_set.
    def enterComponent_parameter_set(self, ctx:McCompParser.Component_parameter_setContext):
        pass

    # Exit a parse tree produced by McCompParser#component_parameter_set.
    def exitComponent_parameter_set(self, ctx:McCompParser.Component_parameter_setContext):
        pass


    # Enter a parse tree produced by McCompParser#component_define_parameters.
    def enterComponent_define_parameters(self, ctx:McCompParser.Component_define_parametersContext):
        pass

    # Exit a parse tree produced by McCompParser#component_define_parameters.
    def exitComponent_define_parameters(self, ctx:McCompParser.Component_define_parametersContext):
        pass


    # Enter a parse tree produced by McCompParser#component_set_parameters.
    def enterComponent_set_parameters(self, ctx:McCompParser.Component_set_parametersContext):
        pass

    # Exit a parse tree produced by McCompParser#component_set_parameters.
    def exitComponent_set_parameters(self, ctx:McCompParser.Component_set_parametersContext):
        pass


    # Enter a parse tree produced by McCompParser#component_out_parameters.
    def enterComponent_out_parameters(self, ctx:McCompParser.Component_out_parametersContext):
        pass

    # Exit a parse tree produced by McCompParser#component_out_parameters.
    def exitComponent_out_parameters(self, ctx:McCompParser.Component_out_parametersContext):
        pass


    # Enter a parse tree produced by McCompParser#component_parameters.
    def enterComponent_parameters(self, ctx:McCompParser.Component_parametersContext):
        pass

    # Exit a parse tree produced by McCompParser#component_parameters.
    def exitComponent_parameters(self, ctx:McCompParser.Component_parametersContext):
        pass


    # Enter a parse tree produced by McCompParser#ComponentParameterDouble.
    def enterComponentParameterDouble(self, ctx:McCompParser.ComponentParameterDoubleContext):
        pass

    # Exit a parse tree produced by McCompParser#ComponentParameterDouble.
    def exitComponentParameterDouble(self, ctx:McCompParser.ComponentParameterDoubleContext):
        pass


    # Enter a parse tree produced by McCompParser#ComponentParameterInteger.
    def enterComponentParameterInteger(self, ctx:McCompParser.ComponentParameterIntegerContext):
        pass

    # Exit a parse tree produced by McCompParser#ComponentParameterInteger.
    def exitComponentParameterInteger(self, ctx:McCompParser.ComponentParameterIntegerContext):
        pass


    # Enter a parse tree produced by McCompParser#ComponentParameterString.
    def enterComponentParameterString(self, ctx:McCompParser.ComponentParameterStringContext):
        pass

    # Exit a parse tree produced by McCompParser#ComponentParameterString.
    def exitComponentParameterString(self, ctx:McCompParser.ComponentParameterStringContext):
        pass


    # Enter a parse tree produced by McCompParser#ComponentParameterVector.
    def enterComponentParameterVector(self, ctx:McCompParser.ComponentParameterVectorContext):
        pass

    # Exit a parse tree produced by McCompParser#ComponentParameterVector.
    def exitComponentParameterVector(self, ctx:McCompParser.ComponentParameterVectorContext):
        pass


    # Enter a parse tree produced by McCompParser#ComponentParameterSymbol.
    def enterComponentParameterSymbol(self, ctx:McCompParser.ComponentParameterSymbolContext):
        pass

    # Exit a parse tree produced by McCompParser#ComponentParameterSymbol.
    def exitComponentParameterSymbol(self, ctx:McCompParser.ComponentParameterSymbolContext):
        pass


    # Enter a parse tree produced by McCompParser#ComponentParameterDoubleArray.
    def enterComponentParameterDoubleArray(self, ctx:McCompParser.ComponentParameterDoubleArrayContext):
        pass

    # Exit a parse tree produced by McCompParser#ComponentParameterDoubleArray.
    def exitComponentParameterDoubleArray(self, ctx:McCompParser.ComponentParameterDoubleArrayContext):
        pass


    # Enter a parse tree produced by McCompParser#ComponentParameterIntegerArray.
    def enterComponentParameterIntegerArray(self, ctx:McCompParser.ComponentParameterIntegerArrayContext):
        pass

    # Exit a parse tree produced by McCompParser#ComponentParameterIntegerArray.
    def exitComponentParameterIntegerArray(self, ctx:McCompParser.ComponentParameterIntegerArrayContext):
        pass


    # Enter a parse tree produced by McCompParser#ShareBlock.
    def enterShareBlock(self, ctx:McCompParser.ShareBlockContext):
        pass

    # Exit a parse tree produced by McCompParser#ShareBlock.
    def exitShareBlock(self, ctx:McCompParser.ShareBlockContext):
        pass


    # Enter a parse tree produced by McCompParser#ShareBlockCopy.
    def enterShareBlockCopy(self, ctx:McCompParser.ShareBlockCopyContext):
        pass

    # Exit a parse tree produced by McCompParser#ShareBlockCopy.
    def exitShareBlockCopy(self, ctx:McCompParser.ShareBlockCopyContext):
        pass


    # Enter a parse tree produced by McCompParser#DisplayBlock.
    def enterDisplayBlock(self, ctx:McCompParser.DisplayBlockContext):
        pass

    # Exit a parse tree produced by McCompParser#DisplayBlock.
    def exitDisplayBlock(self, ctx:McCompParser.DisplayBlockContext):
        pass


    # Enter a parse tree produced by McCompParser#DisplayBlockCopy.
    def enterDisplayBlockCopy(self, ctx:McCompParser.DisplayBlockCopyContext):
        pass

    # Exit a parse tree produced by McCompParser#DisplayBlockCopy.
    def exitDisplayBlockCopy(self, ctx:McCompParser.DisplayBlockCopyContext):
        pass


    # Enter a parse tree produced by McCompParser#component_ref.
    def enterComponent_ref(self, ctx:McCompParser.Component_refContext):
        pass

    # Exit a parse tree produced by McCompParser#component_ref.
    def exitComponent_ref(self, ctx:McCompParser.Component_refContext):
        pass


    # Enter a parse tree produced by McCompParser#coords.
    def enterCoords(self, ctx:McCompParser.CoordsContext):
        pass

    # Exit a parse tree produced by McCompParser#coords.
    def exitCoords(self, ctx:McCompParser.CoordsContext):
        pass


    # Enter a parse tree produced by McCompParser#reference.
    def enterReference(self, ctx:McCompParser.ReferenceContext):
        pass

    # Exit a parse tree produced by McCompParser#reference.
    def exitReference(self, ctx:McCompParser.ReferenceContext):
        pass


    # Enter a parse tree produced by McCompParser#dependency.
    def enterDependency(self, ctx:McCompParser.DependencyContext):
        pass

    # Exit a parse tree produced by McCompParser#dependency.
    def exitDependency(self, ctx:McCompParser.DependencyContext):
        pass


    # Enter a parse tree produced by McCompParser#DeclareBlock.
    def enterDeclareBlock(self, ctx:McCompParser.DeclareBlockContext):
        pass

    # Exit a parse tree produced by McCompParser#DeclareBlock.
    def exitDeclareBlock(self, ctx:McCompParser.DeclareBlockContext):
        pass


    # Enter a parse tree produced by McCompParser#DeclareBlockCopy.
    def enterDeclareBlockCopy(self, ctx:McCompParser.DeclareBlockCopyContext):
        pass

    # Exit a parse tree produced by McCompParser#DeclareBlockCopy.
    def exitDeclareBlockCopy(self, ctx:McCompParser.DeclareBlockCopyContext):
        pass


    # Enter a parse tree produced by McCompParser#uservars.
    def enterUservars(self, ctx:McCompParser.UservarsContext):
        pass

    # Exit a parse tree produced by McCompParser#uservars.
    def exitUservars(self, ctx:McCompParser.UservarsContext):
        pass


    # Enter a parse tree produced by McCompParser#InitializeBlock.
    def enterInitializeBlock(self, ctx:McCompParser.InitializeBlockContext):
        pass

    # Exit a parse tree produced by McCompParser#InitializeBlock.
    def exitInitializeBlock(self, ctx:McCompParser.InitializeBlockContext):
        pass


    # Enter a parse tree produced by McCompParser#InitializeBlockCopy.
    def enterInitializeBlockCopy(self, ctx:McCompParser.InitializeBlockCopyContext):
        pass

    # Exit a parse tree produced by McCompParser#InitializeBlockCopy.
    def exitInitializeBlockCopy(self, ctx:McCompParser.InitializeBlockCopyContext):
        pass


    # Enter a parse tree produced by McCompParser#SaveBlock.
    def enterSaveBlock(self, ctx:McCompParser.SaveBlockContext):
        pass

    # Exit a parse tree produced by McCompParser#SaveBlock.
    def exitSaveBlock(self, ctx:McCompParser.SaveBlockContext):
        pass


    # Enter a parse tree produced by McCompParser#SaveBlockCopy.
    def enterSaveBlockCopy(self, ctx:McCompParser.SaveBlockCopyContext):
        pass

    # Exit a parse tree produced by McCompParser#SaveBlockCopy.
    def exitSaveBlockCopy(self, ctx:McCompParser.SaveBlockCopyContext):
        pass


    # Enter a parse tree produced by McCompParser#FinallyBlock.
    def enterFinallyBlock(self, ctx:McCompParser.FinallyBlockContext):
        pass

    # Exit a parse tree produced by McCompParser#FinallyBlock.
    def exitFinallyBlock(self, ctx:McCompParser.FinallyBlockContext):
        pass


    # Enter a parse tree produced by McCompParser#FinallyBlockCopy.
    def enterFinallyBlockCopy(self, ctx:McCompParser.FinallyBlockCopyContext):
        pass

    # Exit a parse tree produced by McCompParser#FinallyBlockCopy.
    def exitFinallyBlockCopy(self, ctx:McCompParser.FinallyBlockCopyContext):
        pass


    # Enter a parse tree produced by McCompParser#metadata.
    def enterMetadata(self, ctx:McCompParser.MetadataContext):
        pass

    # Exit a parse tree produced by McCompParser#metadata.
    def exitMetadata(self, ctx:McCompParser.MetadataContext):
        pass


    # Enter a parse tree produced by McCompParser#category.
    def enterCategory(self, ctx:McCompParser.CategoryContext):
        pass

    # Exit a parse tree produced by McCompParser#category.
    def exitCategory(self, ctx:McCompParser.CategoryContext):
        pass


    # Enter a parse tree produced by McCompParser#initializerlist.
    def enterInitializerlist(self, ctx:McCompParser.InitializerlistContext):
        pass

    # Exit a parse tree produced by McCompParser#initializerlist.
    def exitInitializerlist(self, ctx:McCompParser.InitializerlistContext):
        pass


    # Enter a parse tree produced by McCompParser#assignment.
    def enterAssignment(self, ctx:McCompParser.AssignmentContext):
        pass

    # Exit a parse tree produced by McCompParser#assignment.
    def exitAssignment(self, ctx:McCompParser.AssignmentContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionBinaryLess.
    def enterExpressionBinaryLess(self, ctx:McCompParser.ExpressionBinaryLessContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionBinaryLess.
    def exitExpressionBinaryLess(self, ctx:McCompParser.ExpressionBinaryLessContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionBinaryGreater.
    def enterExpressionBinaryGreater(self, ctx:McCompParser.ExpressionBinaryGreaterContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionBinaryGreater.
    def exitExpressionBinaryGreater(self, ctx:McCompParser.ExpressionBinaryGreaterContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionGrouping.
    def enterExpressionGrouping(self, ctx:McCompParser.ExpressionGroupingContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionGrouping.
    def exitExpressionGrouping(self, ctx:McCompParser.ExpressionGroupingContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionBinaryLessEqual.
    def enterExpressionBinaryLessEqual(self, ctx:McCompParser.ExpressionBinaryLessEqualContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionBinaryLessEqual.
    def exitExpressionBinaryLessEqual(self, ctx:McCompParser.ExpressionBinaryLessEqualContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionArrayAccess.
    def enterExpressionArrayAccess(self, ctx:McCompParser.ExpressionArrayAccessContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionArrayAccess.
    def exitExpressionArrayAccess(self, ctx:McCompParser.ExpressionArrayAccessContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionBinaryLogic.
    def enterExpressionBinaryLogic(self, ctx:McCompParser.ExpressionBinaryLogicContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionBinaryLogic.
    def exitExpressionBinaryLogic(self, ctx:McCompParser.ExpressionBinaryLogicContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionInteger.
    def enterExpressionInteger(self, ctx:McCompParser.ExpressionIntegerContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionInteger.
    def exitExpressionInteger(self, ctx:McCompParser.ExpressionIntegerContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionExponentiation.
    def enterExpressionExponentiation(self, ctx:McCompParser.ExpressionExponentiationContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionExponentiation.
    def exitExpressionExponentiation(self, ctx:McCompParser.ExpressionExponentiationContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionBinaryGreaterEqual.
    def enterExpressionBinaryGreaterEqual(self, ctx:McCompParser.ExpressionBinaryGreaterEqualContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionBinaryGreaterEqual.
    def exitExpressionBinaryGreaterEqual(self, ctx:McCompParser.ExpressionBinaryGreaterEqualContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionZero.
    def enterExpressionZero(self, ctx:McCompParser.ExpressionZeroContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionZero.
    def exitExpressionZero(self, ctx:McCompParser.ExpressionZeroContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionMyself.
    def enterExpressionMyself(self, ctx:McCompParser.ExpressionMyselfContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionMyself.
    def exitExpressionMyself(self, ctx:McCompParser.ExpressionMyselfContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionUnaryPM.
    def enterExpressionUnaryPM(self, ctx:McCompParser.ExpressionUnaryPMContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionUnaryPM.
    def exitExpressionUnaryPM(self, ctx:McCompParser.ExpressionUnaryPMContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionPrevious.
    def enterExpressionPrevious(self, ctx:McCompParser.ExpressionPreviousContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionPrevious.
    def exitExpressionPrevious(self, ctx:McCompParser.ExpressionPreviousContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionTrinaryLogic.
    def enterExpressionTrinaryLogic(self, ctx:McCompParser.ExpressionTrinaryLogicContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionTrinaryLogic.
    def exitExpressionTrinaryLogic(self, ctx:McCompParser.ExpressionTrinaryLogicContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionFloat.
    def enterExpressionFloat(self, ctx:McCompParser.ExpressionFloatContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionFloat.
    def exitExpressionFloat(self, ctx:McCompParser.ExpressionFloatContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionPointerAccess.
    def enterExpressionPointerAccess(self, ctx:McCompParser.ExpressionPointerAccessContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionPointerAccess.
    def exitExpressionPointerAccess(self, ctx:McCompParser.ExpressionPointerAccessContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionIdentifier.
    def enterExpressionIdentifier(self, ctx:McCompParser.ExpressionIdentifierContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionIdentifier.
    def exitExpressionIdentifier(self, ctx:McCompParser.ExpressionIdentifierContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionBinaryEqual.
    def enterExpressionBinaryEqual(self, ctx:McCompParser.ExpressionBinaryEqualContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionBinaryEqual.
    def exitExpressionBinaryEqual(self, ctx:McCompParser.ExpressionBinaryEqualContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionBinaryPM.
    def enterExpressionBinaryPM(self, ctx:McCompParser.ExpressionBinaryPMContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionBinaryPM.
    def exitExpressionBinaryPM(self, ctx:McCompParser.ExpressionBinaryPMContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionUnaryLogic.
    def enterExpressionUnaryLogic(self, ctx:McCompParser.ExpressionUnaryLogicContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionUnaryLogic.
    def exitExpressionUnaryLogic(self, ctx:McCompParser.ExpressionUnaryLogicContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionStructAccess.
    def enterExpressionStructAccess(self, ctx:McCompParser.ExpressionStructAccessContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionStructAccess.
    def exitExpressionStructAccess(self, ctx:McCompParser.ExpressionStructAccessContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionFunctionCall.
    def enterExpressionFunctionCall(self, ctx:McCompParser.ExpressionFunctionCallContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionFunctionCall.
    def exitExpressionFunctionCall(self, ctx:McCompParser.ExpressionFunctionCallContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionBinaryMD.
    def enterExpressionBinaryMD(self, ctx:McCompParser.ExpressionBinaryMDContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionBinaryMD.
    def exitExpressionBinaryMD(self, ctx:McCompParser.ExpressionBinaryMDContext):
        pass


    # Enter a parse tree produced by McCompParser#ExpressionString.
    def enterExpressionString(self, ctx:McCompParser.ExpressionStringContext):
        pass

    # Exit a parse tree produced by McCompParser#ExpressionString.
    def exitExpressionString(self, ctx:McCompParser.ExpressionStringContext):
        pass


    # Enter a parse tree produced by McCompParser#shell.
    def enterShell(self, ctx:McCompParser.ShellContext):
        pass

    # Exit a parse tree produced by McCompParser#shell.
    def exitShell(self, ctx:McCompParser.ShellContext):
        pass


    # Enter a parse tree produced by McCompParser#SearchPath.
    def enterSearchPath(self, ctx:McCompParser.SearchPathContext):
        pass

    # Exit a parse tree produced by McCompParser#SearchPath.
    def exitSearchPath(self, ctx:McCompParser.SearchPathContext):
        pass


    # Enter a parse tree produced by McCompParser#SearchShell.
    def enterSearchShell(self, ctx:McCompParser.SearchShellContext):
        pass

    # Exit a parse tree produced by McCompParser#SearchShell.
    def exitSearchShell(self, ctx:McCompParser.SearchShellContext):
        pass


    # Enter a parse tree produced by McCompParser#unparsed_block.
    def enterUnparsed_block(self, ctx:McCompParser.Unparsed_blockContext):
        pass

    # Exit a parse tree produced by McCompParser#unparsed_block.
    def exitUnparsed_block(self, ctx:McCompParser.Unparsed_blockContext):
        pass



del McCompParser


"""Handles compilation of C translated instruments into usable binaries"""
from enum import Flag, auto
from typing import Union
from pathlib import Path
from mccode_antlr.instr import Instr
from mccode_antlr.translators.c import CTargetVisitor
from zenlog import log


class CBinaryTarget:
    class Type(Flag):
        single = auto()
        acc = auto()
        mpi = auto()
        nexus = auto()

    def __init__(self, mpi: bool = False, acc: bool = False, count: int = 1, nexus: bool = False):
        self.count = count
        if mpi and acc:
            self.type = CBinaryTarget.Type.acc | CBinaryTarget.Type.mpi
        elif mpi:
            self.type = CBinaryTarget.Type.mpi
        elif acc:
            self.type = CBinaryTarget.Type.acc
        else:
            self.type = CBinaryTarget.Type.single
        # completely orthogonal, probably a bad idea to include here at all:
        if nexus:
            self.type |= CBinaryTarget.Type.nexus


    @property
    def compiler(self) -> str:
        from mccode_antlr.config import config
        if self.type & CBinaryTarget.Type.acc:
            # acc *or* acc + mpi uses the OpenACC compiler
            return config['acc'].as_str_expanded()
        elif self.type & CBinaryTarget.Type.mpi:
            # mpi (no-acc) uses the OpenMPI compiler
            return config['mpi']['cc'].as_str_expanded()
        # otherwise use the standard C compiler
        return config['cc'].as_str_expanded()

    @property
    def flags(self) -> list[str]:
        from mccode_antlr.config import config
        if self.type & CBinaryTarget.Type.acc:
            # acc *or* acc + mpi uses the OpenACC compiler *and* flags
            return config['flags']['acc'].as_str_expanded().split()
        # otherwise use the standard compiler flags
        return config['flags']['cc'].as_str_expanded().split()

    @property
    def linker_flags(self) -> list[str]:
        from mccode_antlr.config import config
        return config['flags']['ld'].as_str_expanded().split()

    @property
    def extra_flags(self) -> list[str]:
        """Only a little confusing ..."""
        from mccode_antlr.config import config
        extras = []
        if self.type & CBinaryTarget.Type.mpi:
            extras.extend(config['flags']['mpi'].as_str_expanded().split())
        if self.type & CBinaryTarget.Type.nexus:
            extras.extend(config['flags']['nexus'].as_str_expanded().split())
        return extras


def instrument_source(instrument: Instr, generator: dict, config: dict, verbose: bool = None):
    if verbose is None:
        verbose = config.get('verbose', False)
    # TargetVisitor to uses instrument.registries (and the CTargetVisitor *appends* its LIBC_REGISTRY if necessary)
    visitor = CTargetVisitor(instrument, generate=generator, config=config, verbose=verbose)
    # Does the conversion by 'visiting' the instrument, then returns the full string of the generated source code
    return visitor.contents()


def compile_instrument(instrument: Instr, target: CBinaryTarget, output: Union[str, Path] = None,
                       recompile: bool = False, replace: bool = True, dump_source: bool = True, **kwargs):
    from os import R_OK, access
    from subprocess import run, CalledProcessError
    from mccode_antlr.config import config
    log.info(f'Compile {instrument.name}')
    # determine a name and location for the binary file
    if output is None:
        # use the current working directory if nothing provided
        output = Path()
    if not isinstance(output, Path):
        # allow for a string input to be interpreted as a path
        output = Path(output)
    if output.is_dir():
        # allow for the user to specify only the output *directory*
        output = output.joinpath(instrument.name).with_suffix(config['ext'].get(str))

    if output.exists() and not recompile:
        raise RuntimeError(f"Output {output} exists but recompile is not requested.")
    if output.exists() and not replace:
        return output

    log.info(f'Sort out flags for compilation')

    # the type of binary requested determines (some of) the required flags:
    compiler_flags = target.flags + target.extra_flags
    linker_flags = target.linker_flags
    # the instrument-defined flags are always(?) linker flags:
    # the flags in an instrument *might* contain ENV, CMD, GETPATH directives which need to be expanded via decode:
    linker_flags.extend([word for flag in instrument.decoded_flags() for word in flag.split()])

    log.info(f'{compiler_flags = }')
    log.info(f'{linker_flags = }')

    # Why is this addition necessary?
    if any('OPENACC' in word for word in compiler_flags) and any('NeXus' in word for word in compiler_flags):
        compiler_flags.append('-D__GNUC__')

    # The solitary '-' specifies *where* the stdin source should be processed, which is critical for getting
    # linking flags right on (some) Linux systems
    command = [target.compiler, *compiler_flags, '-o', str(output), '-', *linker_flags]
    source = instrument_source(instrument, **kwargs)
    if dump_source:
        source_file = Path().joinpath(output.parts[-1]).with_suffix('.c')
        log.info(f'Source written in {source_file}')
        with open(source_file, 'w') as cfile:
            cfile.write(source)
    # print(f"Compile using {command}")
    result = run(command, input=source, text=True, capture_output=True)
    if result.returncode:
        raise RuntimeError(f"Compilation failed producing output\n{result.stdout}\n  and error\n{result.stderr}")
    #
    # try:
    #     run(command, input=instrument_source(instrument, **kwargs), text=True, check=True)
    # except CalledProcessError as error:
    #     raise RuntimeError(f'Compilation failed, raising error {error}')

    if not output.exists():
        raise RuntimeError(f"Compilation should have produced {output}, but it does not appear to exist")
    if not access(output, R_OK):
        raise RuntimeError(f"{output} exists but is not an executable")

    return output


def run_compiled_instrument(binary: Path, target: CBinaryTarget, options: str, capture=False, dry_run: bool = False):
    from subprocess import run, CalledProcessError
    from platform import system
    from mccode_antlr.config import config

    # If NeXus output is requested and the InstrumentDescriptionFile is needed, run a different script entirely...
    #   TODO think about actually doing this?
    # if target.flags & CBinaryTarget.Type.nexus and idf_required:
    #     run([config['idfgen'].get(str), str(binary), *options])

    command = []
    if target.type & CBinaryTarget.Type.mpi:
        # we execute mpirun
        command.append(config['mpi']['run'].as_str_expanded())
        # which takes optional flags
        if target.count == 0:
            print("Using system default number of mpirun processes")
        else:
            command.extend(['-np', str(target.count)])
        if config['machinefile'].exists():
            command.extend(['-machinefile', config['machinefile'].as_str_expanded()])
        # and requires trickery if we want to restrict the GPU used
        if target.type & CBinaryTarget.Type.acc and 'Windows' != system():
            # Each worker should have the environment variable CUDA_VISIBLE_DEVICES defined as the value of
            # OMPI_COMM_WORLD_LOCAL_RANK, which gets sent by MPI to the worker. This assigment must take place
            # *on* the worker, which requires hijacking the executable that MPI runs
            command.append('acc_gpu_bind')
            raise NotImplementedError('CUDA GPU binding not yet implemented')

    binary = binary.resolve()
    if not binary.exists():
        raise RuntimeError(f'Can not execute {binary} since it does not exist.')

    # In normal operation, the binary is provided with options
    command.extend([str(binary), *options.split()])
    # which we then execute:
    if dry_run:
        log.info(f'Would execute {command}')
        return ""
    result = run(command, capture_output=capture)
    if result.returncode and capture:
        raise RuntimeError(f'Execution of {command} failed with output\n{result.stdout}\n and error\n{result.stderr}')
    elif result.returncode:
        raise RuntimeError(f'Execution of {command} failed, see above for error message(s)')

    return (result.stdout + result.stderr) if capture else ""



# Generated from /home/g/Code/mccode_antlr-antlr/mccode_antlr/grammar/C.g4 by ANTLR 4.13.0
from antlr4 import *
if "." in __name__:
    from .CParser import CParser
else:
    from CParser import CParser

# This class defines a complete generic visitor for a parse tree produced by CParser.

class CVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by CParser#functionCall.
    def visitFunctionCall(self, ctx:CParser.FunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#primaryExpression.
    def visitPrimaryExpression(self, ctx:CParser.PrimaryExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#genericSelection.
    def visitGenericSelection(self, ctx:CParser.GenericSelectionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#genericAssocList.
    def visitGenericAssocList(self, ctx:CParser.GenericAssocListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#genericAssociation.
    def visitGenericAssociation(self, ctx:CParser.GenericAssociationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#postfixExpression.
    def visitPostfixExpression(self, ctx:CParser.PostfixExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#argumentExpressionList.
    def visitArgumentExpressionList(self, ctx:CParser.ArgumentExpressionListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#unaryExpression.
    def visitUnaryExpression(self, ctx:CParser.UnaryExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#unaryOperator.
    def visitUnaryOperator(self, ctx:CParser.UnaryOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#castExpression.
    def visitCastExpression(self, ctx:CParser.CastExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#multiplicativeExpression.
    def visitMultiplicativeExpression(self, ctx:CParser.MultiplicativeExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#additiveExpression.
    def visitAdditiveExpression(self, ctx:CParser.AdditiveExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#shiftExpression.
    def visitShiftExpression(self, ctx:CParser.ShiftExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#relationalExpression.
    def visitRelationalExpression(self, ctx:CParser.RelationalExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#equalityExpression.
    def visitEqualityExpression(self, ctx:CParser.EqualityExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#andExpression.
    def visitAndExpression(self, ctx:CParser.AndExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#exclusiveOrExpression.
    def visitExclusiveOrExpression(self, ctx:CParser.ExclusiveOrExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#inclusiveOrExpression.
    def visitInclusiveOrExpression(self, ctx:CParser.InclusiveOrExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#logicalAndExpression.
    def visitLogicalAndExpression(self, ctx:CParser.LogicalAndExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#logicalOrExpression.
    def visitLogicalOrExpression(self, ctx:CParser.LogicalOrExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#conditionalExpression.
    def visitConditionalExpression(self, ctx:CParser.ConditionalExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#assignmentExpression.
    def visitAssignmentExpression(self, ctx:CParser.AssignmentExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#assignmentOperator.
    def visitAssignmentOperator(self, ctx:CParser.AssignmentOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#expression.
    def visitExpression(self, ctx:CParser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#constantExpression.
    def visitConstantExpression(self, ctx:CParser.ConstantExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#declaration.
    def visitDeclaration(self, ctx:CParser.DeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#declarationSpecifiers.
    def visitDeclarationSpecifiers(self, ctx:CParser.DeclarationSpecifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#declarationSpecifiers2.
    def visitDeclarationSpecifiers2(self, ctx:CParser.DeclarationSpecifiers2Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#declarationSpecifier.
    def visitDeclarationSpecifier(self, ctx:CParser.DeclarationSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#initDeclaratorList.
    def visitInitDeclaratorList(self, ctx:CParser.InitDeclaratorListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#initDeclarator.
    def visitInitDeclarator(self, ctx:CParser.InitDeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#storageClassSpecifier.
    def visitStorageClassSpecifier(self, ctx:CParser.StorageClassSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#typeSpecifier.
    def visitTypeSpecifier(self, ctx:CParser.TypeSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#structOrUnionSpecifier.
    def visitStructOrUnionSpecifier(self, ctx:CParser.StructOrUnionSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#structOrUnion.
    def visitStructOrUnion(self, ctx:CParser.StructOrUnionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#structDeclarationList.
    def visitStructDeclarationList(self, ctx:CParser.StructDeclarationListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#structDeclaration.
    def visitStructDeclaration(self, ctx:CParser.StructDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#specifierQualifierList.
    def visitSpecifierQualifierList(self, ctx:CParser.SpecifierQualifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#structDeclaratorList.
    def visitStructDeclaratorList(self, ctx:CParser.StructDeclaratorListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#structDeclarator.
    def visitStructDeclarator(self, ctx:CParser.StructDeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#enumSpecifier.
    def visitEnumSpecifier(self, ctx:CParser.EnumSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#enumeratorList.
    def visitEnumeratorList(self, ctx:CParser.EnumeratorListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#enumerator.
    def visitEnumerator(self, ctx:CParser.EnumeratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#enumerationConstant.
    def visitEnumerationConstant(self, ctx:CParser.EnumerationConstantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#atomicTypeSpecifier.
    def visitAtomicTypeSpecifier(self, ctx:CParser.AtomicTypeSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#typeQualifier.
    def visitTypeQualifier(self, ctx:CParser.TypeQualifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#functionSpecifier.
    def visitFunctionSpecifier(self, ctx:CParser.FunctionSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#alignmentSpecifier.
    def visitAlignmentSpecifier(self, ctx:CParser.AlignmentSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#declarator.
    def visitDeclarator(self, ctx:CParser.DeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#directDeclarator.
    def visitDirectDeclarator(self, ctx:CParser.DirectDeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#vcSpecificModifer.
    def visitVcSpecificModifer(self, ctx:CParser.VcSpecificModiferContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#gccDeclaratorExtension.
    def visitGccDeclaratorExtension(self, ctx:CParser.GccDeclaratorExtensionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#gccAttributeSpecifier.
    def visitGccAttributeSpecifier(self, ctx:CParser.GccAttributeSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#gccAttributeList.
    def visitGccAttributeList(self, ctx:CParser.GccAttributeListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#gccAttribute.
    def visitGccAttribute(self, ctx:CParser.GccAttributeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#nestedParenthesesBlock.
    def visitNestedParenthesesBlock(self, ctx:CParser.NestedParenthesesBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#pointer.
    def visitPointer(self, ctx:CParser.PointerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#typeQualifierList.
    def visitTypeQualifierList(self, ctx:CParser.TypeQualifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#parameterTypeList.
    def visitParameterTypeList(self, ctx:CParser.ParameterTypeListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#parameterList.
    def visitParameterList(self, ctx:CParser.ParameterListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#parameterDeclaration.
    def visitParameterDeclaration(self, ctx:CParser.ParameterDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#identifierList.
    def visitIdentifierList(self, ctx:CParser.IdentifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#typeName.
    def visitTypeName(self, ctx:CParser.TypeNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#abstractDeclarator.
    def visitAbstractDeclarator(self, ctx:CParser.AbstractDeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#directAbstractDeclarator.
    def visitDirectAbstractDeclarator(self, ctx:CParser.DirectAbstractDeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#typedefName.
    def visitTypedefName(self, ctx:CParser.TypedefNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#initializer.
    def visitInitializer(self, ctx:CParser.InitializerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#initializerList.
    def visitInitializerList(self, ctx:CParser.InitializerListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#designation.
    def visitDesignation(self, ctx:CParser.DesignationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#designatorList.
    def visitDesignatorList(self, ctx:CParser.DesignatorListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#designator.
    def visitDesignator(self, ctx:CParser.DesignatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#staticAssertDeclaration.
    def visitStaticAssertDeclaration(self, ctx:CParser.StaticAssertDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#statement.
    def visitStatement(self, ctx:CParser.StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#labeledStatement.
    def visitLabeledStatement(self, ctx:CParser.LabeledStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#compoundStatement.
    def visitCompoundStatement(self, ctx:CParser.CompoundStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#blockItemList.
    def visitBlockItemList(self, ctx:CParser.BlockItemListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#blockItem.
    def visitBlockItem(self, ctx:CParser.BlockItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#expressionStatement.
    def visitExpressionStatement(self, ctx:CParser.ExpressionStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#selectionStatement.
    def visitSelectionStatement(self, ctx:CParser.SelectionStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#iterationStatement.
    def visitIterationStatement(self, ctx:CParser.IterationStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#multiLineMacroStatement.
    def visitMultiLineMacroStatement(self, ctx:CParser.MultiLineMacroStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#forCondition.
    def visitForCondition(self, ctx:CParser.ForConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#forDeclaration.
    def visitForDeclaration(self, ctx:CParser.ForDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#forExpression.
    def visitForExpression(self, ctx:CParser.ForExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#jumpStatement.
    def visitJumpStatement(self, ctx:CParser.JumpStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#compilationUnit.
    def visitCompilationUnit(self, ctx:CParser.CompilationUnitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#translationUnit.
    def visitTranslationUnit(self, ctx:CParser.TranslationUnitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#externalDeclaration.
    def visitExternalDeclaration(self, ctx:CParser.ExternalDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#functionDefinition.
    def visitFunctionDefinition(self, ctx:CParser.FunctionDefinitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CParser#declarationList.
    def visitDeclarationList(self, ctx:CParser.DeclarationListContext):
        return self.visitChildren(ctx)



del CParser


from dataclasses import dataclass
from .utilities import escape_str_for_c

@dataclass
class RawC:
    filename: str
    line: int
    source: str
    translated: str = None

    def __str__(self):
        return self.source

    @property
    def is_empty(self):
        # this could also check for and exclude comments
        return len(self.source.translate(str.maketrans('', '', ' \t\n'))) == 0

    @staticmethod
    def from_tuple(p: tuple):
        if isinstance(p, tuple) and len(p) == 3 and \
                isinstance(p[0], str) and isinstance(p[1], int) and isinstance(p[2], str):
            return RawC(*p)
        if isinstance(p, tuple) and len(p) == 2 and isinstance(p[0], int) and isinstance(p[1], str):
            return RawC("", *p)
        if isinstance(p, str):
            return RawC("", -1, p)
        raise RuntimeError(f"No conversion to RawC from\n{p}")

    def to_c(self):
        """Use the preprocessor #line directive to aid in debugging produced C source code."""
        # return f'#line {self.line} "{self.filename}"\n{self.source}'
        if self.translated is None:
            from zenlog import log
            log.error('RawC.to_c() called before translation')
        return self.translated

    @property
    def fn(self):
        return escape_str_for_c(self.filename), self.line

    def copy(self):
        return RawC(self.filename, self.line, self.source)

    def __contains__(self, value):
        return value in self.source


def blocks_to_raw_c(*args):
    raw_c = [x if isinstance(x, RawC) else RawC.from_tuple(x) for x in args]
    # Filter out empty blocks
    output = tuple(x for x in raw_c if not x.is_empty)
    return output if len(output) else ()



from dataclasses import dataclass
from ..common import Expr


@dataclass
class Jump:
    target: str
    relative_target: int
    iterate: bool
    condition: Expr
    absolute_target: int = -1

    def to_file(self, output, wrapper):
        # self.target is one of "PREVIOUS", "PREVIOUS_{i}, "MYSELF", "MYSELF_{i}, "NEXT",
        # or "{component instance identifier}"
        jump_name = self.target
        if '_' in jump_name and (jump_name.startswith('PREVIOUS') or jump_name.startswith('NEXT')):
            jump_name = jump_name.split('_')[0]
        if abs(self.relative_target) > 1:
            jump_name += f' ({abs(self.relative_target)})'
        when_iter = 'ITERATE' if self.iterate else 'WHEN'
        print(wrapper.line('JUMP', [jump_name, when_iter, str(self.condition)]), file=output)

    def __contains__(self, value):
        return value in self.condition
